"""Generate input files for EEASiSSS.x and run it."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-17'
__license__ = 'GPLv3+'

from collections import defaultdict
import copy
import logging
import os
from pathlib import Path
import random
import re
import shutil
import subprocess

import fortranformat as ff
import numpy as np

from viperleed.calc.classes.sitetype import Atom_type
from viperleed.calc.lib.leedbase import EV_TO_HARTREE
from viperleed.calc.lib.periodic_table import PERIODIC_TABLE

logger = logging.getLogger(__name__)


###############################################
#                 GLOBALS                     #
###############################################

# Conversion factor Bohr-radii <-> Angstrom
angst_to_bohr = 1.8897259886
bohr_to_angst = 0.529177249


class PSFile:
    """Helper class to read atomic phaseshift files generated by EEASiSSS."""

    def __init__(self, atom, filename, ps_energy_step,
                 delete_file_after_read=True):
        self.file = Path(filename).resolve()
        self.atom = atom
        self.energy_step = ps_energy_step
        self._read(delete_file_after_read)

    def _read(self, delete_file_after_read=True):
        with open(self.file, 'r') as psfile:
            reade = -1.
            pslist = []
            ps_of_e = {}
            # dictionary of pslist for energy, where pslist is a list of
            #  floats as read from the PS-file
            for (line_nr, line) in enumerate(psfile):
                if line_nr == 0:
                    self.firstline = line[2:]  # ignore I2 at beginning
                else:   # line should contain data
                    # put phaseshift floats into values
                    values = [float(s) for s in line.split()]
                    # this takes care of multi-line pshifts
                    if reade < 0 or values[0] == reade + self.energy_step:
                        # new energy value, start phaseshift value list
                        if reade >= 0:
                            ps_of_e[reade] = pslist
                        reade = values[0]
                        pslist = values[1:]
                    else:
                        pslist.extend(values)
            ps_of_e[reade] = pslist
        self.energies = sorted(list(ps_of_e.keys()))
        self.phaseshifts = [ps_of_e[en] for en in self.energies]

        if delete_file_after_read:
            os.remove(self.file)


def runPhaseshiftGen_old(sl, rp,
                     psgensource='eeasisss',
                     excosource='seSernelius',
                     atdenssource='atom_density_files'):
    """Creates required input for eeasisss, then runs it. Reads the output
    files and extracts information for PHASESHIFTS file, then returns that
    information (without writing PHASESHIFTS)."""

    if rp.source_dir is None:
        raise RuntimeError("No tensorleed source directory specified")

    # Since there's a limit on the path length, pick the shortest one
    # we can find around. Use './' and '../' shorthands if necessary.
    # Notice that we use os.path.rel_path and not Path.relative_to as
    # the latter does not produce './' and '../' shorthands
    shortpath = str(rp.source_dir)
    if len(os.path.relpath(rp.source_dir)) < len(shortpath):
        shortpath = os.path.relpath(rp.source_dir, start=rp.workdir)

    manual_copy = len(shortpath) > 62
    if manual_copy:
        # Copy over files. Charge densities will be copied later
        # in _collect_eeasisss_input when we know which elements
        shutil.copy2(Path(shortpath, excosource), excosource)
        shortpath = '.'

    # Here it is important NOT TO .resolve() excosource that will
    # end up in the input file. Otherwise all the mess above with
    # deciding what's the shortest possible way is useless
    psgensource = Path(rp.source_dir, psgensource)
    excosource = Path(shortpath, excosource)

    if os.name == 'nt':
        psgensource = psgensource.with_suffix('.exe')                           # TODO: does this cover it or should we use 'win' in sys.platform()?

    if not psgensource.is_file():
        raise FileNotFoundError('Could not find PHASESHIFTS executable at '
                                f'{psgensource}. Did you forget to compile it? '
                                'Try running make in the tensorleed directory')

    _, lmax = rp.get_limits('LMAX')
    nsl, newbulkats = sl.with_extra_bulk_units(rp, 1)

    # The following originally calculated phaseshifts first for only the bulk,
    #   then for the slab, taking phaseshifts from the bulk calculations if the
    #   site is present at all in the bulk. However, this means that different
    #   muffin tin parameters are applied for the different phaseshifts.
    # New solution:
    #   Take phaseshifts from the slab calculation only. Average only over
    #   atoms NOT added as "new bulk" ("newbulkats" list above) to avoid
    #   influence of the (false) bottom "surface".
#    for bulk in [True,False]:
#    wsl = bsl if bulk else nsl  # working slab for this iteration
    # before starting on unit cell, determine whether supercell is needed:
    blocks = []  # tuples of sites (with poscar elements) and elements
    #        (same as POSCAR if no ELEMENT_MIX, ELEMENT_MIX elements if not)
    for site in nsl.sitelist:
        if site.el in rp.ELEMENT_MIX:
            for el in rp.ELEMENT_MIX[site.el]:
                blocks.append((site, el))
        else:
            blocks.append((site, site.el))
    supercell_size = 1 # Supercell size
    if len(rp.ELEMENT_MIX) > 0:
        minnum = -1
        for (site, el) in [(site, el) for (site, el) in blocks if site.el
                           in rp.ELEMENT_MIX and (site.occ[el] > 0. or
                                                  el in site.mixedEls)]:
            al = [at for at in nsl if at.site == site]
            atom_count = len(al)*site.occ[el]
            if minnum < 0 or (minnum > atom_count >= 0):
                minnum = atom_count
        # we want at least 2 atoms of each element in each site type:
        if 0 < minnum < 2.0:
            supercell_size = int(np.ceil(2.0/minnum))
        elif minnum == 0:
            supercell_size = 100  # large number, will be decreased below
    if supercell_size > 1:  # some checks to make sure it doesn't get too large
        maxcells = 20  # upper limit on supercell size
        maxats = 500   # upper limit on atoms in supercell
        if supercell_size > maxcells:
            supercell_size = maxcells
            # don't warn - this is a large unit cell either way.
        if nsl.n_atoms * supercell_size > maxats:
            logger.debug(
                "Phaseshift generation: Given element "
                "concentrations would require a very large supercell. "
                "Element concentrations for low-occupancy elements will be "
                "increased to avoid this. This only concerns the phaseshifts "
                "calculation and should not cause problems.")
            # determine minimum size to have 2 of each element
            minsize = 1
            for site in [s for s in nsl.sitelist if s.el in rp.ELEMENT_MIX]:
                ats = len([at for at in nsl if at.site == site])
                els = len([el for el in rp.ELEMENT_MIX[site.el]
                           if site.occ[el] > 0.])
                minsize = max(minsize, int(np.ceil(2*els / ats)))
            supercell_size = max(minsize, int(maxats / nsl.n_atoms))

    if supercell_size > 1:  # construct supercell to get enough atoms
        # Try to get similar number of repeats
        # along the two unit-cell directions
        a_size = int(np.ceil(np.sqrt(supercell_size)))
        while supercell_size % a_size:
            a_size += 1
        b_size = supercell_size // a_size
        nsl = nsl.make_supercell(np.diag((a_size, b_size)))

    subatlists = {}     # atlist per block tuple
    for site in nsl.sitelist:
        if site.el in rp.ELEMENT_MIX:
            occdict = {}
            for (k, v) in site.occ.items():
                if v > 0.0 or k in rp.ELEMENT_MIX[site.el]:
                    occdict[k] = v
            # sort by occupancy values
            occdict = dict(sorted(occdict.items(),
                                  key=lambda kv: (kv[1], kv[0])))
            al = [at for at in nsl if at.site == site]
            totats = len(al)
            for el in occdict:
                subatlists[(site.label, el)] = []                               # TODO: once site.__hash__ has been implemented, we switch back from site.label to just site. See commit 40c37096231811226e65efc37ed9781a54e27b87
                reqats = int(np.ceil(totats * site.occ[el]))
                reqats = max(2, reqats)
                while reqats > 0 and len(al) > 0:
                    at = random.choice(al)
                    subatlists[(site.label, el)].append(at)                     # TODO: once site.__hash__ has been implemented, we switch back from site.label to just site. See commit 40c37096231811226e65efc37ed9781a54e27b87
                    al.remove(at)
                    reqats -= 1
            if len(al) > 0:  # should never happen
                logger.warning("Error in PHASESHIFTS file "
                               "generation: Not all atoms were distributed!")
        else:
            subatlists[(site.label, site.el)] = [at for at                      # TODO: once site.__hash__ has been implemented, we switch back from site.label to just site. See commit 40c37096231811226e65efc37ed9781a54e27b87
                                                 in nsl if at.site == site]
    blocks = [(site, el) for (site, el) in blocks
              if len(subatlists[(site.label, el)]) > 0]                         # TODO: once site.__hash__ has been implemented, we switch back from site.label to just site. See commit 40c37096231811226e65efc37ed9781a54e27b87
#    if bulk:
#        bulksites = [site for (site,el) in blocks]
#    else:
#        # site objects are different in the new slab; copy to equivalent
#        oldbulksites = bulksites[:]
#        bulksites = []
#        for oldsite in oldbulksites:
#            for newsite in [site for (site,el) in blocks]:
#                if newsite.isEquivalent(oldsite):
#                    bulksites.append(newsite)
    # start writing output, will be input for EEASiSSS code:
    output = ''
    output += "STRUCTURE:\n"
    output += rp.systemName+" "+rp.timestamp+"\n"

    output += ("'s'  1.00 16      BulkOrSlab('b' or 's'), "
               "LatticeConstant(Angstroms), nshell\n")
    # write transposed unit cell matrix
    for i in range(0, 3):
        ol = ''
        for j in range(0, 3):
            s = str(round(nsl.ucell.transpose()[i, j], 4))+' '
            ol += s.ljust(8)
        if i == 0:
            ol += '      CoordinatesOfUnitCell(LCunits)\n'
        else:
            ol += '\n'
        output += ol

    output += (str(nsl.n_atoms)+"  "+str(nsl.n_atoms)
               + "                  #AtomTypes,#OccupiedAtomTypes\n")
    ptl = [el.lower() for el in PERIODIC_TABLE]

    chemels = {}
    chem_el_paths = {}
    for (site, el) in blocks:
        if el in rp.ELEMENT_RENAME:
            chemel = rp.ELEMENT_RENAME[el]
        elif el.lower() in ptl:
            chemel = el.capitalize()
        else:
            _err = ("Error generating PHASESHIFTS file: Could not "
                    "identify "+el+" as a chemical element. Define "
                    "ELEMENT_RENAME or ELEMENT_MIX parameter.")
            logger.error(_err)
            raise RuntimeError(_err)
        el_charge_density_path = (rp.source_dir / "atom_density_files" /
                                  chemel / (f"chgden{chemel}")).resolve()
        charge_density_short_path = (Path(atdenssource) / chemel /
                                    f"chgden{chemel}")
        os.makedirs(charge_density_short_path.resolve().parent, exist_ok=True)
        shutil.copy2(el_charge_density_path, charge_density_short_path.resolve())
        chem_el_paths[el] = charge_density_short_path.resolve().relative_to(rp.workdir)
        chemels[el] = chemel

    nsl.sort_by_z(bottom_to_top=False)
    for at in nsl:
        # realcartpos = np.dot(nsl.ucell, at.pos)
        # use the "real" cartesian system, with Z going up
        for (site, el) in blocks:
            if at in subatlists[(site.label, el)]:                              # TODO: once site.__hash__ has been implemented, we switch back from site.label to just site. See commit 40c37096231811226e65efc37ed9781a54e27b87
                chemel = chemels[el]
                chgdenpath = chem_el_paths[el]
        output += ("1 "+str(PERIODIC_TABLE.index(chemel)+1)
                   + ".  0.  0.  '"+str(chgdenpath)+"'\n")
        ol = ""
        for j in range(0, 3):
            # ol += str(round(realcartpos[j],4))+" "
            ol += str(round(at.cartpos[j], 4))+" "
        output += ol + "     Coordinates(LCunits)\n"
    output += "SCATTERING: \n"
    output += "'"+str(excosource)+"' |exchange-correlation file\n"
    output += str(lmax)+"  |lmax\n"
    output += "'r' |SelectCalculation: 'relativistic'/'nonrelativistic'\n"
    output += "'p' |SelectOutput: 'phaseshift'/'sigma'/'dataflow'\n"
    output += "'n' |phaseshift: print_SpinPhaseShift? 'yes'/no'\n"
    output += ("'n' |phaseshift,sigma: print_log10(DsigmaDomega)? "
               "'yes'/'no'\n")
    output += "'n' |phaseshift,sigma: print_DsigmaDtheta? 'yes'/'no'\n"
    output += "'n' |phaseshift,sigma: print_Sherman? 'yes'/'no'\n"
    output += ("'n' |phaseshift,sigma: print_TotalCrossSection? "
               "'yes'/'no'\n")
    output += "'n' |dataflow: print_RhoPot? 'yes'/'no'\n"
    output += "'n' |dataflow: print_PSvsE? 'yes'/'no'\n"
    output += "'n' |dataflow: print_WaveFunction? 'yes'/'no'\n"

    # Energy step used for phaseshift calculation eeasisss.
    # Does not need to match theory energy step as phaseshifts will be interpolated anyways.
    ps_energy_step = max(1.0, round(rp.THEO_ENERGIES.step))
    output += ("0.0," + str(round(rp.THEO_ENERGIES.max + 20, 1)) # hardcoded lower boundary of 0
               + ","+str(ps_energy_step)
               + " |'phaseshift'/'dataflow' run: E1->E2,PS_Estep\n")
    output += "100.,100.,1  |'sigma' run: E1->E2,NumVals\n"
    output += "MT OPTIMIZATION (Nelder-Mead):\n"
    output += "'*'           | NMselect: '*' or '+'\n"
    output += "0.05d0   | NMlambda, simplex point shift.\n"
    output += "3            | NMiter, no. of simplex calls.\n"
    output += "1.d-04    ! NMeps,  simplex epsilon.\n"
    output += "0.125d0 | NMepsit, epsilon iteration, eps=eps*epsit.\n"
#    outfilename = 'eeasisss-input-bulk' if bulk else 'eeasisss-input-slab'
    eeasisss_input_path = rp.workdir / 'eeasisss-input'
    try:
        with open(eeasisss_input_path, 'w') as wf:
            wf.write(output)
    except Exception:
        logger.error("Phaseshift data generation: Failed to write "
                     f"{eeasisss_input_path}. Proceeding with execution...")
    # if os.name == 'nt':
    #     logger.error("Phaseshift generation is currently not "
    #                  "supported on Windows. Use a linux shell to run "
    #                  "phaseshift generation script.")
    #     raise EnvironmentError("Phaseshift generation is currently not "
    #                            "supported on Windows.")
    # else:

    phaseshifts_log_name = f"phaseshifts-{rp.timestamp}"
    phaseshifts_log_path = (rp.workdir / phaseshifts_log_name).with_suffix(".log")

    # RUNS phaseshift program
    ps_output = subprocess.run(psgensource,
                               cwd=rp.workdir,
                               input=output,
                               encoding='ascii',
                               capture_output=True)

    # Write EEASISSS output to log file
    try:
        with open(phaseshifts_log_path, 'w') as wf:
            wf.writelines([
                f"Output of EEASISSS called with args '{ps_output.args}'\n",
                f"Exit code: {ps_output.returncode}\n",
                "stdout:\n",
                f"{ps_output.stdout}",
                ])
            if ps_output.stderr:
                wf.writelines(["stderr:", f"{ps_output.stderr}"])
    except OSError:
        logger.error("Could not write EEASISSS stdout/stderr to log file. "
                     "Execution will proceed, but this may indicate a permission "
                     "error.")
    else:
        logger.debug(f"EEASISSS stdout/stderr saved to {phaseshifts_log_path}.")

    # go through all the files that were generated by EEASiSSS and read
    filelist = [filename for filename in os.listdir(rp.workdir) if
                filename.startswith('PS.r.')]
    rgx = re.compile(r'PS\.r\.[0-9]+\.[0-9]+')
    remlist = []
    for filename in filelist:
        m = rgx.match(filename)
        if not m:
            remlist.append(filename)
        else:
            if m.group(0) != filename:
                remlist.append(filename)
            else:
                try:
                    int(filename.split('.')[-1])
                except Exception:
                    remlist.append(filename)
    for filename in remlist:
        filelist.remove(filename)
    if not filelist:
        logger.error("Phaseshift generation failed: No output files found.")
        raise RuntimeError("Phaseshift generation failed.")

    # sort by atom number
    filelist.sort(key=lambda filename: int(filename.split('.')[-1]))
    firstline = "" # first line contains parameters for the potential
    # data from all the PS files
    processed_files = [PSFile(at, rp.workdir/filename, ps_energy_step)
                       for (at, filename) in zip(nsl, filelist)]
    firstline = processed_files[0].firstline

    # make energies into array, check that energies are the same for all atoms
    phaseshift_energies = np.array([psfile.energies
                                    for psfile in processed_files])
    if any(np.var(phaseshift_energies, axis=0) > 1e-6):
        raise RuntimeError("Phaseshift energies are not the same for all atoms.")
    phaseshift_energies = phaseshift_energies[0]

    # make phaseshifts into an array for further processing
    atoms_phaseshifts = np.array([psfile.phaseshifts
                                  for psfile in processed_files])

    # make phaseshifts continuos and wrap the starting value
    atoms_phaseshifts = np.apply_along_axis(adjust_phaseshifts, axis=1,
                                            arr=atoms_phaseshifts)

    # average over atoms of site
    outvals = defaultdict(list)  # This is outvals[energy][block][L]            # TODO: clean up this messy data structure
    atom_index_map = {at: i for i, at in enumerate(nsl)}
    for site, block_element in blocks:
        equivalent_atoms = [at for at in subatlists[(site.label, block_element)]# TODO: once site.__hash__ has been implemented, we switch back from site.label to just site. See commit 40c37096231811226e65efc37ed9781a54e27b87
                            if at not in newbulkats]
        if not equivalent_atoms:
            continue
        equivalent_atoms_ind = [atom_index_map[at]
                                for at in equivalent_atoms]

        # Check variance of phaseshifts for "equivalent" atoms. Warn
        # if it's very large. May help identify averaging of different
        # coordination environments
        equivalent_atoms_ps = atoms_phaseshifts[equivalent_atoms_ind, ...]
        if equivalent_atoms_ps.var(axis=0).max() > 1e-1:                        # TODO: is this a sensible threshold?
            logger.warning("Large variance in phaseshifts for atoms of type "
                           f"{block_element} at site {site}. This may indicate"
                           " averaging of non-equivalent atom environments.")
            rp.setHaltingLevel(2)
        site_averaged_phaseshifts = equivalent_atoms_ps.mean(axis=0)

        phaseshift_element = equivalent_atoms[0].el
        for energy, phaseshifts in zip(phaseshift_energies,
                                       site_averaged_phaseshifts):
            outvals[energy].append(
                (phaseshift_element, block_element, site, phaseshifts.tolist())
                )
    outvals = dict(outvals)

    # clean up
    # bss = "-bulk" if bulk else "-slab"
    bss = ""
    try:
        os.rename('logfile', 'eeasisss-logfile'+bss)
    except Exception:
        logger.warning("Failed to rename phaseshift generation file "
                       "'logfile' to 'eeasisss-logfile"+bss+"'")
    try:
        os.rename('QMTvsE', 'eeasisss-QMTvsE'+bss)
    except Exception:
        logger.warning("Failed to rename phaseshift generation file "
                       "'QMTvsE' to 'eeasisss-QMTvsE"+bss+"'")
    try:
        os.rename('RMTvsE', 'eeasisss-RMTvsE'+bss)
    except Exception:
        logger.warning("Failed to rename phaseshift generation file "
                       "'RMTvsE' to 'eeasisss-RMTvsE"+bss+"'")
    try:
        os.rename('V0vsE', 'eeasisss-V0vsE'+bss)
    except Exception:
        logger.warning("Failed to rename phaseshift generation file "
                       "'V0vsE' to 'eeasisss-V0vsE"+bss+"'")
    # sort blocks in outvals:
    outvalsSorted = {}
    outvalLength = 0
    # Due to a bug in eeasisss, it does not generate exactly the energies it
    # is asked to; this can result in different energies being present for
    # the bulk and the slab calculations. Workaround: Discard energies where
    # not all sites are present.
    for en in outvals:
        outvalsSorted[en] = []
        if len(outvals[en]) > outvalLength:
            outvalLength = len(outvals[en])
    # first sort by POSCAR elements, same order as POSCAR:
    for el in sl.elements:
        if el in rp.ELEMENT_MIX:
            chemelList = rp.ELEMENT_MIX[el]
        else:
            chemelList = [el]
        # then by sites:
        siteList = [site for site in sl.sitelist if site.el == el]
        for cel in chemelList:
            for site in siteList:
                for en in outvalsSorted:
                    for (o_el, o_cel, o_site, pslist) in outvals[en]:
                        if (o_el == el and o_cel == cel
                                and o_site.isEquivalent(site)):
                            outvalsSorted[en].append(pslist)
    # return:
    # writePHASESHIFTS wants values as list of tuples and energies in Hartree:
    phaseshifts = []
    for en in outvalsSorted:
        if len(outvalsSorted[en]) == outvalLength:
            # drop energies where phaseshift was not calculated for all sites
            phaseshifts.append((en*EV_TO_HARTREE, outvalsSorted[en])) # conversion eV to Hartree
    if firstline == "":
        logger.error("Could not find first line for PHASESHIFTS file "
                     "(should contain MUFTIN parameters).")
        firstline = "ERROR: first line not found in eeasisss output\n"
        rp.setHaltingLevel(2)
    else:
        # add number of blocks to firstline
        nblocks = len(phaseshifts[0][1])
        firstline = str(nblocks).rjust(3)+"  "+firstline
        # remove the "PS.r.**.**"
        firstline = re.sub(r"PS\.r\.[0-9]+\.[0-9]+", "", firstline)
    return (firstline, phaseshifts)


def adjust_phaseshifts(phaseshifts_per_energy):
    r"""Make phase-shifts continuous and wraps the starting value.

    Phase-shifts used in the LEED calculation are $\pi$ periodic
    in principle and may be returned with jumps of $\pi$ by the
    generation code. However, the subsequent calculation may
    interpolate the phase shifts and thus may not be able to
    deal with these jumps. This function removes the jumps and
    wraps the first value to the interval $[-\pi/2, \pi/2]$.
    This is done by transforming the input phaseshifts $p_n$
    to the output phaseshifts $p'_n$ by
    $$
    p'_0 = (p_0% + \pi/2) % \pi - \pi/2
    p'_n = p'_{n-1} + (p_n - p_{n-1}) % \pi
         = p'_0 + \sum_{i=1}^{n} (p_i - p_{i-1}) % \pi
    $$

    Parameters
    ----------
    phaseshifts_per_energy : numpy.ndarray
        The energy-dependent phase shifts for one atom to be
        jump-corrected. Shape (N,).

    Returns
    -------
    adjusted_phaseshifts : numpy.ndarray
        A jump-corrected and wrapped version of phaseshifts_per_energy.
        Shape (N,).
    """
    adjusted_phaseshifts = np.empty_like(phaseshifts_per_energy)
    adjusted_phaseshifts[0] = (phaseshifts_per_energy[0] + np.pi/2) % np.pi
    adjusted_phaseshifts[0] -= np.pi/2

    ps_deltas_mod_pi = (np.diff(phaseshifts_per_energy) + np.pi/2) % np.pi
    ps_deltas_mod_pi -= np.pi/2
    ps_deltas_sum_mod_pi = np.cumsum(ps_deltas_mod_pi)

    adjusted_phaseshifts[1:] = adjusted_phaseshifts[0] + ps_deltas_sum_mod_pi
    return adjusted_phaseshifts


def runPhaseshiftGen(sl, rp, psgensource=os.path.join('tensorleed', 'eeasisss_new', 'eeas')):
    """
    Creates the input for EEASISSS, runs it and read the output from the generated files.
    Returns phaseshift information (but does not yet write PHASSHIFTS file) in addition with the firstline string which
    contains the muffin tin potential paramenters.
    """

    ###############################################
    #                 Settings                    #
    ###############################################

    # l_max always set to 18 (not expensive)
    l_max = 18

    S_overlap = rp.S_OVL
    additional_layers = 5  # variable ?

    input_file_name = "EEASISSS-input.txt"
    log_filename = "EEASISSS-log.txt"  # log file

    # subdirectory with phaseshifts
    ps_outdir = 'PS_out'
    remove_outdir = False # INFO: useful for debugging

    # Energy grid paramenters
    E2 = round(rp.THEO_ENERGIES.max + 20, 2)  # add 20 eV to energy range
    Estep = round(rp.THEO_ENERGIES.step, 2) # may need a lower limit of 1.0 like the old eeasisss version

    ###############################################
    #                 Start code                  #
    ###############################################

    # Take atoms from slab; build supercell matching the approximate mixed element concentrations; order into atom_types
    atom_types, nsl, uct = make_atom_types(rp, sl,
                                           additional_layers)  # uct is unit cell vectors, nsl is slab with added bulk

    atom_pos_block, mt_params_block = make_atoms_input_blocks(atom_types, additional_layers, S_overlap)

    # The unction below takes care of formatting the input so it can be used with eeasisss
    # calls function to format and organize atoms
    input_file_txt = format_eeasisss_input(atom_types, uct, atom_pos_block, mt_params_block, l_max, E2, Estep, rp)

    # Write input file
    try:
        with open(input_file_name, 'w') as wf:
            wf.write(input_file_txt)
    except Exception:
        logger.error("Phaseshift data generation: Failed to write "
                     + input_file_name + ". Proceeding with execution...")

    ###############################################
    # Call EEASISSS with input file
    ###############################################

    psgensource = os.path.join('tensorleed', 'eeasisss_new', 'eeasisss')
    psgensource = os.path.join(rp.source_dir, psgensource) # otherwise the location would not be known
    atlib_dir = os.path.join('tensorleed', 'eeasisss_new', 'atlib/') # atom density files, by Sernelius
    atlib_dir = os.path.join(rp.source_dir, atlib_dir)
    outdir_path = os.path.join(".",ps_outdir+"/")
    # create directory for individual phaseshift files if not yet present
    os.makedirs(outdir_path, exist_ok=True)

    # execution of EEASISSS requires the executable eeasisss to be copied to work directory
    eeasisss_exec_path = os.path.join('tensorleed', 'eeasisss_new', 'eeasisss')
    eeasisss_exec_path = os.path.join(rp.source_dir, eeasisss_exec_path)
    try:
        shutil.copy2(eeasisss_exec_path, rp.workdir)
    except Exception:
        logger.error("Could not copy file eeasisss required by EEASISSS. Phaseshift generation will fail if not present.")
        rp.setHaltinglevel(2)

    # We are now ready to call EEASISS
    psgencommand = [psgensource, '-i', input_file_name, '-l', log_filename, '-a', atlib_dir, '-o', outdir_path]
    logger.debug("Now calling EEASISSS...")
    try:
        complete = subprocess.run(psgencommand, capture_output=True, text=True)  # returns Subprocess.complete instance
        complete.check_returncode()  # If returncode is non-zero, raise a CalledProcessError. -> except & finally
        logger.debug("EEASISSS execution finished without error. See EEASISSS-log.txt") # -> only if returncode == 0
    except Exception:  # can subprocess even fail?
        logger.error("Error during EEASISSS execution.")
        raise RuntimeError("Subprocess EEASISSS failed")
    finally:
        if complete.stdout != "":
            logger.debug("EEASISSS stdout:\n" + str(complete.stdout))
        if complete.stderr != "":
            logger.error("EEASISSS stderr:\n" + str(complete.stderr)) # Put error from Fortran into log!

    #when done, remove eeasisss executable from work directory again:
    try:
        os.remove('eeasisss')
    except Exception:
        logger.warning("Could not remove eeasisss executable from work directory.") # Not a big deal if this happens.

    # Now the results of the phaseshift calculation are read out and formatted for further processing
    firstline, phaseshifts = convert_eeasisss_output(sl, rp, atom_types, l_max, E2, Estep, ps_outdir)

    # Finish up by moving relevant files to work directory and then removing the out_dir
    move_EEASISSS_files(ps_outdir, log_filename, remove_outdir, Vxc0_files = ['Vxc0Einc', 'Vxc0EincAprx', 'Vxc0EincAprx_v0coef'])

    return (firstline, phaseshifts)


def make_atom_types(rp, sl, additional_layers):
    """Takes slab object and performs various manipulations to produce and sort into atom types.
    To satisfy element mixing, a supercell is created, where sites are occupied roughly matching the expected element
    ratios. The atoms from this extended cell are sorted by site, element bulk layer (if applicable).
    A number of additional bulk layers are added below the surface cell for phaseshift generation. These will be given
    converging muffin tin boundary radii, to fulfill a theoretical requirement."""

    # sl = original slab, nsl = extended

    extended_cell, _ = sl.with_extra_bulk_units(rp, additional_layers)
    extended_cell.project_c_to_z() # project C to Z for phaseshifts only
    extended_cell.collapse_cartesian_coordinates()
    uct = extended_cell.ucell.transpose() # unit cell vectors in matrix
    nsl = extended_cell

    # lowest z position from original slab: (anything below is new bulk) !!! LEED coordinates
    max_z_sl = max([atom.cartpos[2] for atom in sl])                            # TODO: Issue #174

    blocks = []
    #        (same as POSCAR if no ELEMENT_MIX, ELEMENT_MIX elements if not)
    for site in nsl.sitelist:
        if site.el in rp.ELEMENT_MIX:
            for el in rp.ELEMENT_MIX[site.el]:
                blocks.append((site, el))
        else:
            blocks.append((site, site.el))
    scsize = 1  # Super cell size
    if len(rp.ELEMENT_MIX) > 0:
        minnum = -1
        for (site, el) in [(site, el) for (site, el) in blocks if site.el
                                                                  in rp.ELEMENT_MIX and (site.occ[el] > 0. or
                                                                                         el in site.mixedEls)]:
            al = [at for at in nsl if at.site == site]
            atcount = len(al) * site.occ[el]
            if minnum < 0 or (minnum > atcount >= 0):
                minnum = atcount
        # we want at least 2 atoms of each element in each site type:
        if 0 < minnum < 2.0:
            scsize = int(np.ceil(2.0 / minnum))
        elif minnum == 0:
            scsize = 100  # large number, will be decreased below
    if scsize > 1:  # some checks to make sure it doesn't get too large
        maxcells = 20  # upper limit on supercell size
        maxats = 500  # upper limit on atoms in supercell
        if scsize > maxcells:
            scsize = maxcells
            # don't warn - this is a large unit cell either way.
        if nsl.n_atoms * scsize > maxats:
            logger.debug(
                "Phaseshift generation: Given element "
                "concentrations would require a very large supercell. "
                "Element concentrations for low-occupancy elements will be "
                "increased to avoid this. This only concerns the phaseshifts "
                "calculation and should not cause problems.")
            # determine minimum size to have 2 of each element
            minsize = 1
            for site in [s for s in nsl.sitelist if s.el in rp.ELEMENT_MIX]:
                ats = len([at for at in nsl if at.site == site])
                els = len([el for el in rp.ELEMENT_MIX[site.el]
                           if site.occ[el] > 0.])
                minsize = max(minsize, int(np.ceil(2 * els / ats)))
            scsize = max(minsize, int(maxats / nsl.n_atoms))


    if scsize > 1:  # construct supercell to get enough atoms                   # TODO: code repeated from runPhaseshiftGen_old
        # Try to get similar number of repeats
        # along the two unit-cell directions
        a_size = int(np.ceil(np.sqrt(scsize)))
        while scsize % a_size:
            a_size += 1
        b_size = scsize // a_size
        nsl = nsl.make_supercell(np.diag((a_size, b_size)))

    # Write new unit cell vectors; to be used for input
    uct = nsl.ucell.transpose()

    nsl.update_cartesian_from_fractional()
    # Cell is now fully expanded and has right size

    # Generate dict with nearest neighbour distances, used to determine MT radii !!
    # Also used as atom list instead of nsl.atlist
    NN_dict = nsl.get_nearest_neighbours()
    nsl.update_cartesian_from_fractional()

    # Here we introduce the atom_types dict, which will be very important going forward
    atom_types = {}
    atom_types_in_bulk = {}  # bulk layers added into this dict for now, will be combined later

    # Go through sites, decide if element mix is necessary. If so, take care of it
    for site in nsl.sitelist:
        if site.el in rp.ELEMENT_MIX:
            occdict = {}
            for (k, v) in site.occ.items():
                if v > 0.0 or k in rp.ELEMENT_MIX[site.el]:
                    occdict[k] = v
            # sort by occupancy values
            occdict = dict(sorted(occdict.items(),
                                  key=lambda kv: (kv[1], kv[0])))
            al = [atom for atom in NN_dict.keys() if atom.site == site]
            totats = len(al)
            for el in occdict:
                reqats = int(np.ceil(totats * site.occ[el]))
                reqats = max(2, reqats)
                while reqats > 0 and len(al) > 0:
                    atom = random.choice(al)
                    new_bulk = True if atom.cartpos[2] > max_z_sl else False    # TODO: Issue #174
                    NN_dist = NN_dict[atom]
                    if not new_bulk:
                        if (site, el, new_bulk) not in atom_types.keys():
                            atom_types[(site, el, new_bulk)] = Atom_type(el, str(site), new_bulk)
                        atom_types[(site, el, new_bulk)].add_atom(atom, NN_dist)
                    else:
                        if (site, el, new_bulk) not in atom_types_in_bulk.keys():
                            atom_types_in_bulk[(site, el, new_bulk)] = Atom_type(el, str(site), new_bulk)
                        atom_types_in_bulk[(site, el, new_bulk)].add_atom(atom, NN_dist)
                    al.remove(atom)
                    reqats -= 1
            if len(al) > 0:  # should never happen
                logger.warning("Error in PHASESHIFTS file "
                               "generation: Not all atoms were distributed!")
        else:
            for atom in NN_dict.keys():
                if atom.site == site:
                    new_bulk = True if atom.cartpos[2] > max_z_sl else False    # TODO: Issue #174
                    NN_dist = NN_dict[atom]
                    if not new_bulk:
                        if (atom.site, atom.el, new_bulk) not in atom_types.keys():
                            atom_types[(atom.site, atom.el, new_bulk)] = Atom_type(atom.el, str(atom.site), new_bulk)
                        atom_types[(atom.site, atom.el, new_bulk)].add_atom(atom, NN_dist)
                    else:
                        if (atom.site, atom.el, new_bulk) not in atom_types_in_bulk.keys():
                            atom_types_in_bulk[(atom.site, atom.el, new_bulk)] = Atom_type(atom.el,
                                                                                                               str(atom.site),
                                                                                                               new_bulk)
                        atom_types_in_bulk[(atom.site, atom.el, new_bulk)].add_atom(atom, NN_dist)


    # Now simplify by remapping atom_types to a new dict with unique integer ID...
    re_map = {}
    for type_id, old_ref in enumerate(atom_types.keys(),
                                      1):  # old_ref is (sublayer_id, atom.site, atom.el, new_bulk)
        re_map[old_ref] = type_id
    atom_types = {re_map[id]: atom_types[id] for id in re_map}  # list comprehension magic


    types_to_add = {} # bulk layers (by type) to be added to atom_types
    for (site, el, new_bulk) in atom_types_in_bulk:
        for atom in atom_types_in_bulk[(site, el, new_bulk)].atoms:
            NN_dist = atom_types_in_bulk[(site, el, new_bulk)].smallest_NN_dist
            layer = estimate_bulk_layer(atom, nsl, max_z_sl, additional_layers)
            if (site, el, layer) not in types_to_add.keys():
                types_to_add[(site, el, layer)] = Atom_type(el, str(site), new_bulk, layer)
            types_to_add[(site, el, layer)].add_atom(atom, NN_dist)

    # Finally, add the new bulk to atom_types
    for i in range(additional_layers+1):
        for key in types_to_add.keys():
            if i == key[2]:
                type_id += 1
                atom_types[type_id] = types_to_add[key]
            else:
                continue
    return atom_types, nsl, uct


def format_eeasisss_input(atom_types, uct, atom_pos_block, mt_block, l_max, E2, Estep, rp):
    """
    Produces the input file for EEASISSS (version from 2021).
    """
    ##############################
    # Create EEASiSSS input
    ##############################
    logger.debug("Creating EEASISSS input")
    input_file_header = ""  # Header with info about project
    input_file_header += "STRUCTURE:\n"
    input_file_header += "! " + rp.systemName + " " + rp.timestamp + "\n"
    input_file_header += str(os.getcwd()) + "\n"
    #######
    # Unit cell block
    #######
    input_file_uc_block = ""  # unit cell block to be written into input file
    # Unit of length
    input_file_length_line = '1.889727\t!UnitOfLength conversion to Bohr radii\n'  # EEAS converts to Bohr radii internally, conversion factor bohr-Angstrom
    # and thus needs to be given a conversion factor
    # below adapted from old code

    for i in range(0, 3):
        ol = ''
        for j in range(0, 3):
            s = '{: .6f}'.format(uct[i, j]) + '\t'
            ol += s
        ol += '          !CoordinatesOfUnitCell(UOL)\n'
        input_file_uc_block += ol

    ############################
    # Organize atoms and create input block with atom positions
    ############################



    ##############################
    # Options block
    ##############################

    # energy range
    E2_str = str(E2)
    Estep_str = str(Estep)


    input_options = ""
    input_options += "OPTIONS:" + '\n'
    input_options += "'s' !crystal: 'bulk'/'slab'." + '\n'
    input_options += "'n' !'yes'/'no': SpinUp&Down PhaseShifts calc?\"" + '\n'
    input_options += "'n' !'yes'/'no': Rho print?\"" + '\n'
    input_options += "'n' !'yes'/'no': Pot print?\"" + '\n'
    input_options += "'n' !'yes'/'no': WaveFunction print?\"" + '\n'
    input_options += " 000.00  " + E2_str + "    " + Estep_str + " !energy interval E1,E2,Estep" + '\n'
    input_options += "   " + str(1) + "   " + str(l_max) + "                !nthread,lmax" + '\n' # nthread set to 1 since multithreading is disabled for EEASISSS anyways
    input_options += "  1.d-06 1.d-09         !relerr abserr" + '\n'    # I hope these are reasonable values?
    input_options += " DIFFERENTIAL EVOLUTION METHOD" + '\n'
    # TODO check params below
    input_options += "    0.80   0.50         !F_XC,CR_XC" + '\n'       # does this need changing?
    input_options += "    0   1   0           !method" + '\n'           # what does this actually do?
    input_options += "    2   0.800000        !strategy,F_CR" + '\n'
    input_options += " 10000    0             !itermax"                 # fix itermax and other params?

    ##############################
    # Assemble input file
    ##############################
    input_file_txt = ""
    input_file_txt += input_file_header
    input_file_txt += input_file_length_line
    input_file_txt += input_file_uc_block
    input_file_txt += str(len(atom_types)) + '          ! # inequivalent atoms\n'
    input_file_txt += atom_pos_block
    input_file_txt += mt_block
    input_file_txt += input_options

    return input_file_txt


# TODO depreacted recently, replaced by make_atom_types
def organize_atoms_by_types(newbulkats, nsl, sl, rp, additional_layers):
    """
    Takes atoms from the slab (nsl) and groups them for the EEASISSS input. Atoms need to be organized into groups of
    same element and site. Note that a site could have mixed occupation.
    Returns atom_types, which contains the groups of atom types mapped to an index required for identification. Further,
    the input blocks containing the atom positions are produced here.
    """
    atom_types = {} # dict will contain Atom types

    number_of_atoms_in_bulk_layer = len(newbulkats)
    extended_slab, new_bulk_atoms = sl.with_extra_bulk_units(rp, additional_layers)
    for atom in extended_slab:
        if atom not in new_bulk_atoms:
            new_bulk = False
            if (atom.site, atom.el, new_bulk) not in atom_types.keys():
                atom_types[(atom.site, atom.el, new_bulk)] = Atom_type(atom.el, str(atom.site), new_bulk)
                atom_types[(atom.site, atom.el, new_bulk)].add_atom(atom)
            else:
                atom_types[(atom.site, atom.el, new_bulk)].add_atom(atom)
        else:
            continue

    # Now simplify by remapping atom_types to a new dict with unique integer ID...
    re_map = {}
    for type_id, old_ref in enumerate(atom_types.keys(),
                                      1):  # old_ref is (sublayer_id, atom.site, atom.el, new_bulk)
        re_map[old_ref] = type_id
    atom_types = {re_map[id]: atom_types[id] for id in re_map}  # list comprehension magic

    # now add a couple of bulk layers to input
    for i in range(additional_layers):
        atoms_add = new_bulk_atoms[(i+0)*number_of_atoms_in_bulk_layer:(i+1)*number_of_atoms_in_bulk_layer]
        types_to_add = {}
        new_bulk = True
        for atom in atoms_add:
            if (atom.site, atom.el) not in types_to_add.keys():
                types_to_add[(atom.site, atom.el)] = Atom_type(atom.el, str(atom.site), new_bulk)
                types_to_add[(atom.site, atom.el)].add_atom(atom)
            else:
                types_to_add[(atom.site, atom.el)].add_atom(atom)
        for key in types_to_add.keys():
            type_id += 1
            atom_types[type_id] = types_to_add[key]

    return atom_types


def estimate_bulk_layer(atom, nsl, max_z_sl, additional_layers):                # TODO: Issue #174
    """Returns the bulk layer the atom belongs to based on the z coordinate."""
    max_z_nsl = max([atom.cartpos[2] for atom in nsl])
    bulk_layer_thickness = (max_z_nsl - max_z_sl) / additional_layers
    if atom.cartpos[2] < max_z_sl:
        layer = None
    else:
        layer = int(np.ceil((atom.cartpos[2] - max_z_sl) / bulk_layer_thickness))
    return layer


# TODO was deprecated for a while
def organize_atoms_by_sublayers(newbulkats, nsl):
    """
    Unused and outdated.
    Does same thing as organize_atoms_by_type but additionally groups them by sublayers as produced by
    slab.createSublayer(...)
    Produces worse results and EEASISSS takes much longer to execute.
    """

    # iterate over all types of atoms (i.e. all atoms that can be in each site types)
    #create the sublayers
    nsl.create_sublayers(eps=0.001)
    atom_types = {}
    for sublayer_id, sublayer in enumerate(nsl.sublayers, 1):  # enumerate(..., 1) makes sublayer_id start at 1
        for atom in sublayer:
            new_bulk = True if atom in newbulkats else False
            if (sublayer_id, atom.site, atom.el, new_bulk) not in atom_types.keys():
                atom_types[(sublayer_id, atom.site, atom.el, new_bulk)] = Atom_type(atom.el, str(atom.site), new_bulk)
            atom_types[(sublayer_id, atom.site, atom.el, new_bulk)].add_atom(atom)
    # We need to go through all this trouble of making Atom types, since we need to group the atoms into sublayers, but
    # we can't mix elements (which could otherwise happen if we have mixed occupation)
    # Now simplify by remapping atom_types to a new dict with unique integer ID...
    re_map = {}
    for type_id, old_ref in enumerate(atom_types.keys(), 1):  # old_ref is (sublayer_id, atom.site, atom.el, new_bulk)
        re_map[old_ref] = type_id
    atom_types = {re_map[id]: atom_types[id] for id in re_map}  # list comprehension magic

    return atom_types


def make_atoms_input_blocks(atom_types, bulk_layers, S_ovl):
    # Initialize
    input_file_atom_pos_block = ""
    input_file_mt_params_block = ""

    for type_id in atom_types.keys():  # should be ordered

        atom_type = atom_types[type_id]

        # determine rmtmin & rmtmax
        NN_dist = atom_type.get_type_NN_dist()
        if NN_dist > 1e9:

            raise RuntimeError("Nearest Neighbour assignment failed for atom type " + str(atom_type.label))
        layer = atom_type.get_layer()
        # If atom in new bulk, a layer should have been specified
        if atom_type.new_bulk and atom_type.get_layer() == None:
            raise ValueError("Bulk atom type has no layer specified.")

        if not atom_type.new_bulk:
            rmtmin = 0.3*NN_dist
            rmtmax = 0.9*NN_dist
        else:
            # factor that starts small and approaches 1 the closer it is to the last layer
            layer_factor = layer / bulk_layers
            rmtmin = 0.3*NN_dist*(1-layer_factor) + 0.499*NN_dist*layer_factor
            rmtmax = 0.9*NN_dist*(1-layer_factor) + 0.501*NN_dist*layer_factor

        # Very important – convert from Angstrom to Bohr Units!
        rmtmin *= angst_to_bohr
        rmtmax *= angst_to_bohr

        overlap = S_ovl
        element = atom_type.el
        fxc = atom_type.fxc


        type_header = '\t'.join([str(len(atom_type.atoms)), str(atom_type.get_atomic_number()), str(type_id),
                                 str(type_id), '{: .4f}'.format(rmtmin),
                                 '{: .4f}'.format(rmtmax), '{: .4f}'.format(overlap), element])
        if atom_type.new_bulk == True:  # for easy visibility
            type_header += '\t !atom in new bulk'
        input_file_atom_pos_block += type_header + '\n'
        atom_coords_table = ""
        for atom in atom_type.atoms:
            # rescale atomic coordinates from fractional to absolute
            # for this multiply fractional positon by unit cell matrix
            #fractional_vec = np.array([pos for pos in atom.pos])
            #position_vec = uct.dot(fractional_vec)

            atom_coords_table += ' '.join(['{: .5f}'.format(pos) for pos in atom.cartpos])
            atom_coords_table += '\n'
        input_file_atom_pos_block += atom_coords_table

        input_file_mt_params_block += ' '.join([str(type_id), '{: .4f}'.format(rmtmin),
                                                '{: .4f}'.format(rmtmax), '{: .4f}'.format(overlap),
                                                '{: .4f}'.format(fxc), element])
        if type_id == 1:  # only in first line
            input_file_mt_params_block += '\t!iA rmtmin rmtmax rmtS fxc elem'
        elif type_id == 2:  # only in second line
            input_file_mt_params_block += '\t!etc ...'
        input_file_mt_params_block += '\n'
    return input_file_atom_pos_block, input_file_mt_params_block


def convert_eeasisss_output(sl, rp, atom_types, lmax, Emax, Estep, ps_outdir):
    ##############################
    # Read and convert EEASiSSS ouput
    ##############################
    """
    EEASiSSS produces a log and files called PS.xx.n.txt where xx is atomic number and n is an index. These files
    contain phase shifts in radiants for the atom types.
    """

    # go through all the files that were generated by EEASiSSS and read
    filelist = [filename for filename in os.listdir('./' + ps_outdir) if
                filename.startswith('PS.')]

    # Sanity check that output files are as expected:
    expected_files =[]
    for type_id, at_type in atom_types.items():
        expected_filename = 'PS.' + str(at_type.atomic_number) + '.' + str(type_id) + '.txt'
        expected_files.append(expected_filename)
    if set(filelist) == set(expected_files):
        pass  # should be the case...
    else:
        logger.error("Unexpected phaseshift files were generated.") # should never happen if input formatted right
        raise RuntimeError("Phaseshift generation failed.")

    ###################
    # Now we work on reshuffling
    sites_to_average = {} # likely an inelegant solution...
    for type_id, at_type in atom_types.items():
        if at_type.new_bulk:  # kick out all atoms that were added to the new bulk
            continue
        site_el = (at_type.label, at_type.el) # average phaseshifts over all atoms of same element in same site type
        if site_el not in sites_to_average.keys():
            sites_to_average[site_el] = []

        # lmax, E2 and Estep were read and set in format_eeasisss_input(...)
        filename = 'PS.' + str(at_type.atomic_number) + '.' + str(type_id) + '.txt'
        dir_prefix = './'+ ps_outdir +'/'
        # get number of lines: # could also be calculated from energy range
        num_lines = sum(1 for line in open(dir_prefix + filename, 'r', encoding='utf-8'))
        Emin = 0
        ps = np.zeros([num_lines, lmax])  # initialize np array with zeros
        energy = np.zeros([num_lines])
        with open(dir_prefix + filename, 'r', encoding='utf-8') as psfile:
            psfile.readline()  # firstline contains V0 parameters; discard read later
            for j, line in enumerate(psfile):
                values = line.split()
                energy[j] = values[0]
                ps[j, :] = values[1:lmax+1]

            sites_to_average[site_el].append(ps) # move the phaseshifts read from file to dict

    # Now average phaseshifts for each element and site:
    phaseshift_averages = {}
    for (label, el), values in sites_to_average.items():
        stack = np.stack(values, axis=-1)  # stack ps into np array
        ps_avg = np.mean(stack, axis=-1)  # take average; np has got to be the most efficient option.

        phaseshift_averages[(label, el)] = ps_avg

    # read inner potential coefficients from file; see Docstring and Wiki
    coef_filename = 'Vxc0EincAprx_v0coef'
    c0, c1, c2, c3 = read_V0_coefficients(os.path.join(ps_outdir,coef_filename))

    # old output format
    # sort first by element, then by site
    # first sort by POSCAR elements, same order as POSCAR:

    phaseshifts = []
    for j in range(int((Emax - Emin) / Estep) + 1):
        #old output format
        ps = []
        for el in sl.elements:
            if el in rp.ELEMENT_MIX:
                chemelList = rp.ELEMENT_MIX[el]
            else:
                chemelList = [el]
            # then by sites:
            siteList = [site for site in sl.sitelist if site.el == el]
            for cel in chemelList:
                for site in siteList:
                    label = cel + '_in_' + site.label
                    ps.append(phaseshift_averages[(label, cel)][j,:].tolist())

        energy_hartree= energy[j]*EV_TO_HARTREE # conversion from eV to Hartree
        phaseshifts.append([energy_hartree,ps])

    # format into old output format – int at beginning of line is skipped in old version too!
    formater = ff.FortranRecordWriter('(4F8.2)')
    firstline = formater.write([c0,c1,c2,c3])
    firstline += "\t" + "PS.r.00.00\txxxx 210930-000000"  # TODO clean?
    # copied from above, old formatting of first line... TODO
    if firstline == "":
        logger.error("Could not find first line for PHASESHIFTS file "
                     "(should contain MUFTIN parameters).")
        firstline = "ERROR: first line not found in eeasisss output\n"
        rp.setHaltingLevel(2)
    else:
        # add number of blocks to firstline
        nblocks = len(phaseshifts[0][1])
        firstline = str(nblocks).rjust(3) + " " + firstline # technically fails if more then 999 blocks - non-issue?
        # remove the "PS.r.**.**"
        firstline = re.sub(r"PS\.r\.[0-9]+\.[0-9]+", "", firstline)


    return firstline, phaseshifts


def read_V0_coefficients(coef_file_path='PS_out/Vxc0EincAprx_v0coef'):
    """
    Reads coefficients of the innerpotenial from file generated by EEASISSS. This used to be done by taking the first
    line of one of the PS.xx.xx files. While that is still possible, it is a) more elegant to read from the file
    intended for this purpose and b) the coefficients are written to that file with more significant digits.
    Attention: Order of coefficients has changed in the 2021 EEASISSS version. (0123 <-> 1230)
    """
    try:
        #open file
        with open(coef_file_path) as coef_file:
            # file should just be one line with content:
            # v0coef = a b c d
            # where a,b,c,d, are floats coefficient values
            line = coef_file.readline()
            c1, c2, c3, c0 = line.split()[2:6]
        (c1, c2, c3, c0) = [float(c) for c in (c1, c2, c3, c0)]
    except Exception:
        logger.error("Could not open and read file " + coef_file_path)
        raise RuntimeError("Unable to read inner potential coefficients")
    return c0,c1,c2,c3


def move_EEASISSS_files(ps_outdir, log_filename, remove_outdir, Vxc0_files = ['Vxc0Einc', 'Vxc0EincAprx', 'Vxc0EincAprx_v0coef']):
    """
    Moves files from PS_out folder to work folder and then removes PS_out.
    """
    outdir_path = os.path.join(".", ps_outdir + "/")
    # move log file
    shutil.move(os.path.join(outdir_path, log_filename), os.path.join(".", log_filename))
    # move muffin tin potential files
    for file in Vxc0_files:
        shutil.move(os.path.join(outdir_path, file), os.path.join(".", file))
    # only remaining files are individual phaseshift files; can be discarded since they are combines in PHASESHIFTS
    # remove PS_out folder with contents
    if remove_outdir:
        shutil.rmtree(outdir_path)

    # remove files ulog*, udat* and uinp* created by eeasisss programme
    files_in_directory = os.listdir(".")
    filtered_files = [file for file in files_in_directory if (file.startswith('ulog')
                                                     or file.startswith('udat') or file.startswith('uinp'))]
    for file in filtered_files:
        os.remove(os.path.join(".", file))


def count_atoms(types_dict):
    """Counts number of atoms in atom_types dict. Used for debugging."""
    n = 0
    for key in types_dict.keys():
        n += len(types_dict[key].atoms)

    return n


def compare_atoms(types_dict1, types_dict2):
    """Compares two atom type dicts and returns atoms unique to each. Used for debugging."""
    atoms_list1 = []
    for key in types_dict1.keys():
        for atom in types_dict1[key].atoms:
            atoms_list1.append(atom)

    atoms_list2 = []
    for key in types_dict2.keys():
        for atom in types_dict2[key].atoms:
            atoms_list2.append(atom)

    only1 = [atom for atom in atoms_list1 if atom not in atoms_list2]
    only2 = [atom for atom in atoms_list2 if atom not in atoms_list1]

    return only1, only2



"""Package sections of viperleed.calc.

This package contains the main functionality that runs the various
logically different parts of TensErLEED, and takes care of using the
output produced by TensErLEED to modify all the pPthon objects used
for managing the state of a calculation.

Modules
-------
calc_section:
    Contains base definitions, especially which sections are available.
cleanup:
    The (python-only) section that takes care of sorting the mess of
    files generated by TensErLEED after a calculation has finished.
deltas:
    Generate delta-amplitudes, i.e., the perturbative effect on the
    scattered beam amplitudes due to a (small) generalized displacement.
errorcalc:
    Estimate the uncertainty on fit parameters by "displacing" atoms
    only along one "direction".
fd_optimization:
    Optimize one of the parameters that cannot be optimized using the
    tensor-LEED perturbative approximation.
initialization:
    Read and process user input
refcalc:
    Run a full-dynamic (ie, including multiple scattering) calculation
    if the complex amplitudes and intensities of scattered beams.
run_sections:
    Manage the mechanics of running subsequent sections as well as
    cleaning up after each one of them.
search:
    Run a tensor-LEED-based optimization of generalized atomic
    displacements to best fit the calculated I(V) curves to
    those measured experimentally.
superpos:
    Process the output of search or errorcalc to generate I(V) curves
    of configuration that are displaced with respect to the one in
    refcalc.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-11'
__license__ = 'GPLv3+'



"""Module iodeltas of viperleed.calc.files.

Defines functions for reading and writing files relevant to the
delta-amplitudes calculation.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import logging
import os
import shutil

import fortranformat as ff
import numpy as np

from viperleed.calc.files.beams import writeAUXBEAMS
from viperleed.calc.lib.version import Version

logger = logging.getLogger(__name__)


def checkDelta(filename, at, el, rp):
    """Checks whether a given delta file corresponds to the requested
    displacements of a given atom. Returns True or False."""
    eps = 1e-4
    fgeo = []                               # found geo disp
    fvib = []                               # found vib disp
    at.mergeDisp(el)
    if el == "vac":
        dgeo = [np.array([0.0, 0.0, 0.0])]  # requested geo disp
        dvib = [0.0]                        # requested vib disp
    else:
        if el in at.disp_geo:
            dgeo = at.disp_geo[el]
        else:
            dgeo = at.disp_geo["all"]
        if el in at.disp_vib:
            dvib = at.disp_vib[el]
        else:
            dvib = at.disp_vib["all"]
    try:
        with open(filename, "r") as rf:
            lines = rf.readlines()
    except FileNotFoundError:
        logger.error("Error reading file "+filename)
        raise
    if rp.TL_VERSION < Version('1.7.0'):  # formatting changed to I5 in version 1.7
        intlen = 3
    else:
        intlen = 5
    try:
        nbeams = int(lines[1][0:intlen])        # no. of beams
        nvar = int(lines[1][2*intlen:3*intlen]) # no. of variations (geo*vib)
    except Exception:
        logger.error("Error parsing file " + filename)
        raise
    if nbeams != len(rp.ivbeams):
        return False
    if nvar != len(dgeo)*len(dvib):
        return False
    beams = []      # read beams from delta file
    atline = 2
    for i in range(atline, len(lines)):  # read and check beams, then break
        try:        # read hk of beams formatted as F10.5
            fl = [float(s) for s in [lines[i][j:j+10]
                                     for j in range(0, len(lines[i]) - 10,
                                                    10)]]
        except Exception:
            logger.error("Error parsing file "+filename)
            raise
        for j in range(0, int(len(fl)/2)):
            beams.append((fl[2*j], fl[(2*j) + 1]))
        if len(beams) == nbeams:
            atline = i+1
            break
    for (i, hk) in enumerate(beams):   # check beams
        if not rp.ivbeams[i].isEqual_hk(hk, eps=eps):
            return False
    atline += 1   # skip the line after beams
    # geo displacements start here
    rf74x10 = ff.FortranRecordReader("10F7.4")
    parselist = []
    repeats = False
    endgeo = False
    entrycount = 0
    for i in range(atline, len(lines)):
        try:
            fl = [f for f in rf74x10.read(lines[i]) if f is not None]
        except Exception:
            logger.error("Error parsing file "+filename)
            raise
        if len(fl) < 10:
            atline = i+1
            endgeo = True  # short line -> end of geo block
        parselist = parselist + fl
        while len(parselist) >= 3:
            v = parselist[:3]
            new = np.array([v[1], v[2], v[0]])
            parselist = parselist[3:]
            entrycount += 1
            if not repeats:
                append = True
                if fgeo and np.linalg.norm(new - fgeo[0]) < eps:
                    repeats = True
                    append = False
                if append:
                    fgeo.append(new)
                if len(fgeo) == nvar:
                    atline = i+1
                    endgeo = True
                    break
            elif not endgeo:
                if (entrycount-1) % len(fgeo) == 0:  # should repeat here
                    if np.linalg.norm(new - fgeo[0]) > eps:
                        atline = i
                        endgeo = True
                        break
        if endgeo:
            break
    ngeo = len(fgeo)
    # check geometry
    if ngeo != len(dgeo):
        return False
    for (i, gd) in enumerate(fgeo):
        if np.linalg.norm(gd - dgeo[i]) > eps:
            return False
    # vib displacement starts here
    nvib = nvar / ngeo
    if int(nvib) - nvib > 1e-4:
        logger.error("Error reading file "+filename+": number of geometry "
                     "variations found does not match header.")
        return False
    nvib = int(nvib)
    for i in range(atline, len(lines)):
        try:
            fl = [f for f in rf74x10.read(lines[i]) if f is not None]
        except Exception:
            logger.error("Error parsing file "+filename)
            raise
        parselist = parselist + fl
        while len(parselist) >= ngeo:
            fvib.append(parselist[0])
            if any([f != 0. for f in parselist[1:ngeo]]):
                logger.warning("File "+filename+": Found unexpected entries "
                               "in list of vibrational displacements.")
            parselist = parselist[ngeo:]
        if len(fvib) >= nvib:
            break
    # check vibrations:
    if el.lower() == "vac":
        voff = 0.
    else:
        voff = at.site.vibamp[el]
    for (i, f) in enumerate(fvib):
        bv = round(round(dvib[i] + voff, 4) / 0.529177, 4)
        # in bohr radii; rounding twice to account for 1. writing to
        #  delta-input, 2. reading from DELTA file. Precision taken from
        #  TensErLEED GLOBAL
        if abs(f - bv) >= 1e-4:
            return False
    return True


def generateDeltaInput(atom, targetel, sl, rp, deltaBasic="", auxbeams="",
                       phaseshifts=""):
    """
    Generates a PARAM file and delta input for one element of one atom.

    Parameters
    ----------
    atom : Atom
        Atom object for which input should be generated.
    targetel : str
        The element of that atom for which input should be generated. This may
        not be the main atom element.
    sl : Slab
        The Slab object containing atom information.
    rp : Rparams
        The run parameters object.
    deltaBasic : str, optional
        Part of delta input that is the same for all atoms. If not passed,
        will call generateDeltaBasic directly.
    auxbeams : str, optional
        The contents of the AUXBEAMS file. If not passed, will attempt to find
        the AUXBEAMS file and read it.
    phaseshifts : str, optional
        The contents of the PHASESHIFTS file. If not passed, will attempt to
        find the PHASESHIFTS file and read it.

    Returns
    -------
    (str, str, str).
        The delta input, a shortened version of that input for logging, and
        the contents of the required PARAM file.
    """

    if deltaBasic == "":
        deltaBasic = generateDeltaBasic(sl, rp)
    if auxbeams == "":
        # if AUXBEAMS is not in work folder, check SUPP folder
        if not os.path.isfile(os.path.join(".", "AUXBEAMS")):
            if os.path.isfile(os.path.join(".", "SUPP", "AUXBEAMS")):
                try:
                    shutil.copy2(os.path.join(".", "SUPP", "AUXBEAMS"),
                                 "AUXBEAMS")
                except Exception:
                    logger.warning("Failed to copy AUXBEAMS from SUPP folder")
            else:
                logger.warning("generateDeltaInput: AUXBEAMS not found")
        try:
            with open("AUXBEAMS", "r") as rf:
                auxbeams = rf.read()
            if auxbeams[-1] != "\n":
                auxbeams += "\n"
        except Exception:
            logger.error("generateDeltaInput: Could not read AUXBEAMS")
            raise
    if phaseshifts == "":
        try:
            with open("PHASESHIFTS", "r") as rf:
                phaseshifts = rf.read()
            if phaseshifts[-1] != "\n":
                phaseshifts += "\n"
        except Exception:
            logger.error("generateDeltaInput: Could not read PHASESHIFTS")
            raise
    MNLMB = [19, 126, 498, 1463, 3549, 7534, 14484, 25821, 43351, 69322,
             106470, 158067, 227969, 320664, 441320, 595833, 790876, 1033942]
    try:
        beamlist, _, _ = writeAUXBEAMS(ivbeams=rp.ivbeams,
                                       beamlist=rp.beamlist, write=False)
    except Exception:
        logger.error("writeDeltaInput: Exception while getting data from "
                     "writeAUXBEAMS")
        raise

    # merge offsets with displacement lists
    atom.mergeDisp(targetel)

    # generate delta.in
    din = ("""   1                         FORMOUT - 1: formatted output
-------------------------------------------------------------------
--- chemical nature of displaced atom                           ---
-------------------------------------------------------------------
""")
    if targetel.lower() == "vac":
        iel = 0
    else:
        # find number of target element
        i = 0
        for el in sl.elements:
            # this reproduces the order of blocks contained in PHASESHIFTS:
            if el in rp.ELEMENT_MIX:
                chemelList = rp.ELEMENT_MIX[el]
            else:
                chemelList = [el]
            siteList = [s for s in sl.sitelist if s.el == el]
            for cel in chemelList:
                for s in siteList:
                    i += 1
                    if s.isEquivalent(atom.site) and cel == targetel:
                        iel = i
    i4 = ff.FortranRecordWriter("I4")
    f74x3 = ff.FortranRecordWriter('3F7.4')
    ol = i4.write([iel])
    din += ol.ljust(29) + "IEL  - element in PHASESHIFT list"    
    if rp.TL_VERSION < Version('1.7.0'):
        din += """
-------------------------------------------------------------------
--- undisplaced position of atomic site in question             ---
-------------------------------------------------------------------
"""
        ol = f74x3.write([0.0, 0.0, 0.0])
        din += ol.ljust(29) + "CUNDISP - displacement offset"
    din += """
-------------------------------------------------------------------
--- displaced positions of atomic site in question              ---
-------------------------------------------------------------------
"""
    if targetel == "vac":
        geolist = [(0., 0., 0.)]
    elif targetel in atom.disp_geo:
        geolist = atom.disp_geo[targetel]
    else:
        geolist = atom.disp_geo["all"]
    geosteps = len(geolist)
    ol = i4.write([geosteps])
    din += ol.ljust(29) + "NCSTEP - number of displaced positions\n"
    for disp in geolist:
        ol = f74x3.write([disp[2], disp[0], disp[1]])
        din += ol.ljust(29)+"CDISP(z,x,y) - z pointing towards bulk\n"
        # TODO: should we allow this if e.g. HALTING = 1 and just warn instead?
        if any([abs(d) > 1. for d in disp]):
            logger.error(
                "Displacements for delta amplitudes have to be smaller than "
                "one Angstrom! Larger displacements are not reasonable "
                "within the tensor LEED approximation.\n"
                "Found displacement {} for {}.".format(disp, atom))
            raise ValueError("Excessive displacements (>1A) detected.")
    din += (
        """-------------------------------------------------------------------
--- vibrational displacements of atomic site in question        ---
-------------------------------------------------------------------
""")
    if targetel == "vac":
        viblist = [0.]
    elif targetel in atom.disp_vib:
        viblist = atom.disp_vib[targetel]
    else:
        viblist = atom.disp_vib["all"]
    vibsteps = len(viblist)
    ol = i4.write([vibsteps])
    din += ol.ljust(29) + "NDEB - number of vib. amplitudes to be considered\n"
    f74 = ff.FortranRecordWriter("F7.4")
    if targetel.lower() != "vac":
        # "default" vibamp + offset, not just offset
        vibamps = [v + atom.site.vibamp[targetel] for v in viblist]
        if any([v <= 0 for v in vibamps]):
            logger.warning(
                "Vibrational amplitudes for {} contain values <= 0 "
                "(smallest: {:.4f}). Shifting displacement list to avoid "
                "non-positive numbers.".format(atom, min(vibamps)))
            corr = min([v for v in vibamps if v > 0]) - min(vibamps)
            vibamps = [v + corr for v in vibamps]
            for i in range(len(viblist)):
                # can't be done by list comprehension because it should modify
                #   the list that 'viblist' is pointing to, not make a copy
                viblist[i] += corr
    else:
        vibamps = [0.]
    for vibamp in vibamps:
        ol = f74.write([vibamp])
        din += ol.ljust(29)+"DRPER_A\n"

    din_main = deltaBasic + auxbeams + phaseshifts + din
    din_short = deltaBasic + "[AUXBEAMS]\n" + "[PHASESHIFTS]\n" + din

    # write PARAM
    param = ("""C  Parameter statements for delta amplitude calculation, v1.2
C  parameters must be consistent with preceding reference calculation!

C  MLMAX: maximum angular momentum to be considered in calculation
C  MNLMB: number of Clebsh-Gordon coefficients needed in tmatrix() subroutine -
C         set according to current LMAX
"""
             "C         MLMAX:  1  2   3    4    5    6    7     8     9     "
             "10    11     12     13     14     15     16     17     18\n"
             "C         MNLMB: 19 126 498 1463 3549 7534 14484 25821 43351 "
             "69322 106470 158067 227969 320664 441320 595833 790876 1033942\n" """
C  MNPSI: number of phase shift values tabulated in phase shift file
C  MNEL : number of elements for which phase shifts are tabulated
C  MNT0 : number of beams for which delta amplitude calculation is required
C  MNATOMS: currently must be set to 1. In principle number of different atomic
C      positions in a superlattice wrt the reference periodicity when computing
C      TLEED beams for a superlattice not present in the reference structure
C  MNDEB: number of thermal variation steps to be performed (outer var. loop)
C  MNCSTEP: number of geometric variation steps to be performed """
             + "(inner var. loop)\n\n")
    param += "      PARAMETER( MLMAX = {} )\n".format(rp.LMAX.max)
    param += "      PARAMETER( MNLMB = {} )\n".format(MNLMB[rp.LMAX.max-1])
    param += ("      PARAMETER( MNPSI = {}, MNEL = {} )\n"
              .format(len(rp.phaseshifts), (len(rp.phaseshifts[0][1]))))
    param += "      PARAMETER( MNT0 = {} )\n".format(len(beamlist))
    param += "      PARAMETER( MNATOMS = 1 )\n"
    param += "      PARAMETER( MNDEB = {} )\n".format(vibsteps)
    param += "      PARAMETER( MNCSTEP = {} )\n".format(geosteps)
    return din_main, din_short, param


def generateDeltaBasic(sl, rp):
    """Generates the part of the input for delta-amplitudes that is the same
    for all atoms, and returns it as a string."""
    if rp.TL_VERSION < Version('1.7.0'):
        formatter = {'energies': ff.FortranRecordWriter('3F7.2'),
                     'uc': ff.FortranRecordWriter('2F7.4'),
                     'angles': ff.FortranRecordWriter('2F7.4'),
                     }
        lj = 24  # ljust spacing
    else:
        formatter = {'energies': ff.FortranRecordWriter('2F9.2'),
                     'uc': ff.FortranRecordWriter('2F9.4'),
                     'angles': ff.FortranRecordWriter('2F9.4'),
                     'int': ff.FortranRecordWriter('I4'),
                     }
        lj = 30  # ljust spacing
    output = ""
    output += rp.systemName+" "+rp.timestamp+"\n"
    output += (formatter['energies'].write(
        [rp.THEO_ENERGIES.start, rp.THEO_ENERGIES.stop+0.01]).ljust(lj)
        + 'EI,EF\n')
    ucsurf = sl.ab_cell.T
    if sl.bulkslab is None:
        sl.make_bulk_slab(rp)
    ucbulk = sl.bulkslab.ab_cell.T
    output += formatter['uc'].write(ucbulk[0]).ljust(lj) + 'ARA1\n'
    output += formatter['uc'].write(ucbulk[1]).ljust(lj) + 'ARA2\n'
    output += formatter['uc'].write(ucsurf[0]).ljust(lj) + 'ARB1\n'
    output += formatter['uc'].write(ucsurf[1]).ljust(lj) + 'ARB2\n'
    output += (formatter['angles'].write([rp.THETA, rp.PHI]).ljust(lj)
               + 'THETA PHI\n')
    if rp.TL_VERSION >= Version('1.7.0'):
        # TODO: if phaseshifts are calculated differently, change format here
        output += (formatter['int'].write([1]).ljust(lj)
                   + 'PSFORMAT  1: Rundgren_v1.6; 2: Rundgren_v1.7\n')
    return output



"""ViPErLEED poscar utilities."""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-04'
__license__ = 'GPLv3+'

from viperleed.utilities.poscar.cli import PoscarUtilsCLI


PoscarUtilsCLI.run_as_script()



"""Moule domain_params of viperleed.calc.classes.rparams.

Defines the DomainParameters class. Contains of information useful
when a calculation with multiple structural domains is carried out.
This module was originally (2019-06-13) part of the rparams.py module,
refactored by Michele Riva in Oct 2023.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = '2019-2024 ViPErLEED team'
__created__ = '2023-10-23'
__license__ = 'GPLv3+'

from pathlib import Path


class DomainParameters:
    """Stores workdir, slab and Rparams objects for each domain"""

    def __init__(self, workdir, homedir, name):
        # .workdir is the path to the sub-directory for where this
        # domain is calculated, .homedir, the one to the main calc
        # directory
        self.workdir = Path(workdir).resolve()
        self.homedir = Path(homedir).resolve()
        self.name = name  # Domain name as defined by user
        self.sl = None
        self.rp = None
        self.refcalcRequired = False
        self.tensorDir = None



"""Module checker of viperleed.calc.files.parameters.

Defines the ParametersChecker class, useful for checking that parameters
read from a PARAMETERS file do not clash with one another.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-25'
__license__ = 'GPLv3+'

import operator

from .errors import ParameterConflictError


# About the disable: ParametersChecker has only one single job:
# calling its own '_check_xxx' methods. Hence it makes sense to
# have a single entry point (check_parameter_conflicts).
# pylint: disable-next=too-few-public-methods
class ParametersChecker:
    """A container of methods for checking clashes between PARAMETERS.

    To add a new check, it is sufficient to implement a method
    whose name starts with '_check_'. The methods can use the
    self._rpars object to access parameter values.
    """

    def __init__(self):
        """Initialize instance."""
        self._rpars = None

    def check_parameter_conflicts(self, rpars):
        """Make sure the parameters in `rpars` are consistent."""
        self._rpars = rpars
        checker_names = (method_name for method_name in dir(self)
                         if method_name.startswith('_check_'))
        for checker_name in checker_names:
            checker = getattr(self, checker_name)
            checker()

    def _get_attribute_values(self, *attr_names):
        """Yield values of rpars attributes with given names."""
        yield from (getattr(self._rpars, attr) for attr in attr_names)

    def _check_element_name_collisions(self):
        """Make sure ELEMENT_MIX and ELEMENT_RENAME are not in conflict."""
        params = 'ELEMENT_RENAME', 'ELEMENT_MIX'
        values = self._get_attribute_values(*params)
        conflict = operator.and_(*(v.keys() for v in values))
        if conflict:
            message = ('The same POSCAR element cannot appear in both. '
                       'Conflicting elements: ' + ', '.join(conflict))
            raise ParameterConflictError(*params, message=message)

    def _check_and_update_fortran_comp(self):                                   # TODO: it may be necessary to add a .exe on Windows
        """Update FORTRAN_COMP parameter with info from pre and post."""
        # First the 'standard' compiler
        pre, post = self._rpars.FORTRAN_COMP
        if not post and pre in {'ifort', 'gfortran'}:
            self._rpars.getFortranComp(comp=pre)

        # Then MPI
        pre, post = self._rpars.FORTRAN_COMP_MPI
        if not post and pre in {'mpifort', 'mpiifort'}:
            self._rpars.getFortranMpiComp(comp=pre)



"""Module cli of viperleed.utilities.beams.

Defines the main command-line interface for utilities that convert
various experimental beam formats into the EXPBEAMS.csv files used
by ViPErLEED.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-27'
__license__ = 'GPLv3+'


from viperleed.cli_base import ViPErLEEDCLIWithAutoChildren


_HELP = 'convert experimental beams file to EXPBEAMS.csv format'

class BeamsUtilCLI(ViPErLEEDCLIWithAutoChildren,
                   cli_name='beams', help_=_HELP):
    """The main command-line interface for the beams-conversion utility."""

    def __init__(self):
        """Initialize CLI with case-insensitive aliases."""
        super().__init__()
        self.add_child_aliases('from-SATLEED', *_make_aliases('from-SATLEED'))
        self.add_child_aliases('from-TensErLEED',
                               *_make_aliases('from-TensErLEED'),
                               'from-ErLEED',
                               *_make_aliases('from-ErLEED'))


def _make_aliases(name):
    """Return aliases for name."""
    underscore = name.replace('-', '_')
    return name.lower(), underscore, underscore.lower()



"""ViPErLEED utility to convert TensErLEED-style beams to EXPBEAMS."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-12-16'
__license__ = 'GPLv3+'

from viperleed.calc.files.beams import readAUXEXPBEAMS, writeOUTBEAMS
from viperleed.cli_base import ViPErLEEDCLI
from viperleed.utilities.beams import EXPBEAMS_DEFAULT


_DEFAULT_INPUT = 'AUXEXPBEAMS'
_HEADER = '''
This utility reads an AUXEXPBEAMS file (i.e., beams formatted as TensErLEED
experimental input) and writes the contents in the csv formatting applied
in viperleed.calc for THEOBEAMS.csv and EXPBEAMS.csv files.
'''
_HELP = ('convert a TensErLEED-style experimental-'
         'beams input file to ViPErLEED format')


class FromTensErLEEDBeamsCLI(ViPErLEEDCLI,
                             cli_name='from-TensErLEED',
                             help_=_HELP):
    """Utility to convert TensErLEED beams file to EXPBEAMS format."""

    def __call__(self, args=None):
        """Execute utility."""
        # print some info
        print(_HEADER)

        # read the AUXEXPBEAMS file
        filename = ''
        while not filename:
            filename = input('Enter name of TensErLEED-beams file '
                             f'(default: {_DEFAULT_INPUT!r}): ')
            filename = filename or _DEFAULT_INPUT
            try:
                beams = readAUXEXPBEAMS(filename, interactive=True)
            except FileNotFoundError:
                print(f'File {filename} not found.')
                filename = ''

        if not beams:
            print('Error reading AUXEXPBEAMS file.')
            return 1
        # print some info
        print(f'Found file with {len(beams)} beams.\n')

        # relabel beams
        label_width = max(beam.getLabel()[1] for beam in beams)                 # TODO: repeated in many places
        for beam in beams:
            beam.label, _ = beam.getLabel(lwidth=label_width)

        # get output file name
        filename = input('Enter output file name '
                         f'(default: {EXPBEAMS_DEFAULT!r}): ')
        filename = filename or EXPBEAMS_DEFAULT

        # write new file
        try:
            writeOUTBEAMS(beams, filename)
        except OSError:
            print(f'Error writing new file {filename}')
            return 1
        print(f'Wrote output as {filename}')
        return 0


if __name__ == '__main__':
    FromTensErLEEDBeamsCLI.run_as_script()



"""Functions for producing plots of I(V) curves."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2021-10-25'
__license__ = 'GPLv3+'

import logging

import numpy as np

try:
    import matplotlib
except Exception:
    _CAN_PLOT = False
else:
    _CAN_PLOT = True
    matplotlib.rcParams.update({'figure.max_open_warning': 0})
    matplotlib.use('Agg')  # !!! check with Michele if this causes conflicts
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.pyplot as plt
    import matplotlib.ticker as plticker
    plt.style.use('viperleed.calc')

from viperleed.calc.classes.beam import Beam

logger = logging.getLogger(__name__)


def plot_iv(data, filename, labels=[], annotations=[],
            legends=[], formatting={}):
    '''
    Creates a single PDF file containing plots of I(V) curves

    Parameters
    ----------
    data : list of datasets, or single dataset
        dataset : list of Beam or of np.ndarray objects
            contains I(V) data per beam
        All datasets must contain the same beams in the same order, but beams
        can be empty
    filename : str or None
        Name of the file (with or without extension) to which the plots
        will be saved. If None, do not write to file, but return a list
        of matplotlib figure objects instead.
    labels : kwarg, list of str
        Labels for the individual beams, e.g. (h|k) as strings. Must be same
        length as number of beams. If labels are not passed and at least one
        dataset is of type Beam, labels will be generated from that dataset.
    annotations : kwarg, list of str
        Additional information to print for each beam, for example the
        R-factor. Must be same length as the number of beams.
    legends : kwarg, list of str
        How to label the different datasets, in order. Must be same length as
        the number of datasets.
    formatting : kwarg, dict
        dict containing formatting instructions:
        axes : str
            Which axes to draw.
            all:  draw all axes (left, right, top, bottom) for all panels
            lb:   draw only left and bottom axes
            b:    draw only bottom axes
            none: draw no axes except at bottom of last panel in each column
        colors : tuple (str, str)
            Define alternative colors for drawing I(V) curves.
            default: None
        legend : str
            Which legends to print.
            all:   print legend for each panel
            first: print legend only for the first panel on each page
            tr:    print legend only for the top-right panel on each page
            none:  do not print any legend.
        perpage : int or (int, int)
            Define how many figures to plot on each page. Either tuple
            (columns, rows), or single integer (preferred). For single int,
            layout will be adapted automatically. Numbers that are not nicely
            divisible may be rounded up, resulting in some whitespace.
            default: 2 (one column, two rows).

    Returns
    -------
    None

    '''
    global _CAN_PLOT
    if not _CAN_PLOT:
        logger.debug("Necessary modules for plotting not found. Skipping "
                     "R-factor plotting.")
        return

    # check data
    if type(data) not in (list, tuple):
        raise TypeError("Expected data as a list or tuple, found "
                        + str(type(data)))
    if type(data[0]) not in (list, tuple):
        data = [data]       # assume single set of beams
    n_beams = len(data[0])
    for dataset in data:
        if type(dataset) not in (list, tuple):
            raise TypeError(
                "Expected data as a list or tuple, found "
                + str(type(dataset[0])))
        if type(dataset[0]) not in (Beam, np.ndarray):
            raise TypeError(
                "Expected data as a list of viperleed.calc.classes.beam.Beam "
                "or numpy arrays, found " + str(type(dataset[0])))
        elif len(dataset) != n_beams:
            raise ValueError("Different beam sets or not of equivalent "
                             "length.")

    if labels and len(labels) != n_beams:
        raise ValueError("Number of labels does not match number of beams.")
    if annotations and len(annotations) != n_beams:
        raise ValueError("Number of annotations does not match number of "
                         "beams.")

    # set formatting parameters
    if formatting is None:
        formatting = {}
    print_axes = formatting.get('axes', 'all')
    plot_colors = formatting.get('colors', [])
    print_legend = formatting.get('legend', 'all')
    figs_per_page = formatting.get('perpage', 2)
    font_size = formatting.get('font_size', 10)
    line_width = formatting.get('line_width', 1.5)
    curve_line_widths = formatting.get('curve_line_widths',
                                       [line_width]*n_beams)

    # read data
    readlabels = False
    if not labels:
        readlabels = True   # fill labels from first dataset
    xy_per_beam_per_dataset = []
    for dataset in data:
        if type(dataset[0]) == np.ndarray:
            xy_per_beam = [np.copy(xy) for xy in dataset]
        else:    # list of beams
            xy_per_beam = []
            if readlabels:
                if formatting and formatting["overbar"]:
                    labelstyle = "overbar"
                else:
                    labelstyle = "minus"
                labelwidth = max([beam.getLabel(style=labelstyle)[1]
                                  for beam in dataset])
                labels = [beam.getLabel(lwidth=labelwidth, style=labelstyle)[0]
                          for beam in dataset]
                readlabels = False
            for b in dataset:
                xy = np.array([[e, b.intens[e]]
                               for e in sorted(b.intens.keys())])
                xy_per_beam.append(xy)
        xy_per_beam_per_dataset.append(xy_per_beam)

    # find min and max values of x and y for plotting all curves
    # on the same horizontal scale and leaving a little y space for the legend
    all_xy = [xy for xy_per_beam in xy_per_beam_per_dataset
              for xy in xy_per_beam if len(xy) != 0]
    xmin = min(min(xy[:, 0]) for xy in all_xy)
    xmax = max(max(xy[:, 0]) for xy in all_xy)

    ymin = min(min(xy[:, 1]) for xy in all_xy)
    ymax = max(max(xy[:, 1]) for xy in all_xy)
    dy = ymax - ymin

    # Figure out how to arrange the figures
    if type(figs_per_page) == int:
        xfigs = int(np.round(np.sqrt(figs_per_page/2)))
        yfigs = int(np.ceil(figs_per_page / xfigs))
        if xfigs*yfigs != figs_per_page:
            logger.debug("R-factor plots: Figures per page corrected from {} "
                         "to {}".format(figs_per_page, xfigs*yfigs))
            figs_per_page = xfigs*yfigs
    else:
        xfigs, yfigs = figs_per_page
        figs_per_page = xfigs*yfigs
    figsize = (7., 3.5 * yfigs / xfigs)

    # set ticks spacing to 50 eV and round the x limits to a multiple of it
    tick = 50
    oritick = plticker.MultipleLocator(base=tick)
    majortick = oritick
    minortick = None
    xlims = (np.floor(xmin/tick)*tick,
             np.ceil(xmax/tick)*tick)
    dx = xlims[1] - xlims[0]

    # positions and scales, depending on number of figures per page
    figs = []
    gen_scaling = 1 / np.sqrt(xfigs)
    ylims = (ymin - 0.02*dy, ymax + 0.22*dy*font_size/10/gen_scaling)
    namePos = (xlims[0] + 0.02*dx, ylims[1] - 0.1*dy*font_size/10/gen_scaling)  # eg labels
    annotationPos = (namePos[0], namePos[1]-0.085*dy*font_size/10/gen_scaling)  # eg R-factor

    if dx / (tick * gen_scaling) > 16:  # too many labelled ticks
        minortick = majortick
        newbase = int(np.ceil(dx / (tick * gen_scaling * 16))) * tick
        majortick = plticker.MultipleLocator(base=newbase)
    ticklen = 0.3 * font_size * gen_scaling

    # axes helper variables
    axes_visible = {'left': True, 'right': True, 'bottom': True, 'top': True}
    if print_axes != 'all':
        axes_visible['top'] = False
        axes_visible['right'] = False
        if 'l' not in print_axes:
            axes_visible['left'] = False
        if print_axes == 'none':
            axes_visible['bottom'] = False

    # the following will spam the logger with debug messages; disable.
    loglevel = logger.level
    logger.setLevel(logging.INFO)
    try:
        fig_exists = False
        fig_index_on_page = 0
        for ct in range(n_beams):   # iterate through beams
            if all([len(xy_per_beam_per_dataset[i][ct]) == 0
                    for i in range(len(data))]):
                continue   # no data for this beam in any dataset, skip
            if (fig_index_on_page >= figs_per_page) or (not fig_exists):
                # need a new figure
                fig_exists = True # at least one fig exists
                fig_index_on_page = 0
                fig, axs = plt.subplots(yfigs, xfigs, figsize=figsize,
                                        squeeze=False)
                fig.tight_layout()
                axs = axs.flatten(order='C')  # flatten row-style
                fig.subplots_adjust(left=(0.03 / (xfigs * gen_scaling*font_size/10)),
                                    right=(1 - 0.03 / (xfigs * gen_scaling*font_size/10)),
                                    bottom=(0.14 / (yfigs * gen_scaling*font_size/10)),
                                    top=(1 - 0.02 / (yfigs * gen_scaling*font_size/10)),
                                    wspace=0.04 / gen_scaling*font_size/10,
                                    hspace=0.02 / gen_scaling*font_size/10)
                figs.append(fig)
                [ax.set_xlim(*xlims) for ax in axs]
                [ax.set_ylim(*ylims) for ax in axs]
                [ax.get_yaxis().set_visible(False) for ax in axs]
                [sp.set_linewidth(0.5 * line_width * gen_scaling) for ax in axs
                 for sp in ax.spines.values()]
                [ax.xaxis.set_major_locator(majortick) for ax in axs]
                if minortick is not None:
                    [ax.xaxis.set_minor_locator(minortick) for ax in axs]
                for i, ax in enumerate(axs):
                    for k in axes_visible:
                        ax.spines[k].set_visible(axes_visible[k])
                    if (((i//xfigs) + 1 == figs_per_page//xfigs)
                            or n_beams - (ct + i) <= xfigs):
                        ax.set_xlabel("Energy (eV)", fontsize=font_size*gen_scaling,
                                      labelpad=4*gen_scaling)
                        ax.tick_params(
                            which='both', bottom=True,
                            top=axes_visible['top'], axis='x',
                            direction='in', labelsize=0.9*font_size*gen_scaling,
                            pad=0.5*font_size*gen_scaling,
                            width=0.5 * line_width * gen_scaling,
                            length=ticklen)
                        ax.spines['bottom'].set_visible(True)
                    else:
                        ax.tick_params(
                            which='both', bottom=axes_visible['bottom'],
                            top=axes_visible['top'], axis='x', direction='in',
                            labelbottom=False,
                            width=0.5 * line_width * gen_scaling,
                            length=ticklen)
                    if minortick is not None:
                        ax.tick_params(which='minor', length=ticklen*0.5)
            if plot_colors:
                if not all([matplotlib.colors.is_color_like(s)
                            for s in plot_colors]):
                    plot_colors = []
                    logger.warning("plot_iv: Specified colors not "
                                   "recognized, reverting to default colors")
            for i in range(len(data)):
                if i < len(curve_line_widths):
                    lw = curve_line_widths[i] * gen_scaling
                else:
                    lw = curve_line_widths[-1]
                if legends:
                    label = legends[i]
                else:
                    label = 'Beamset {}'.format(i+1)
                xy = xy_per_beam_per_dataset[i][ct]
                if i < len(plot_colors):
                    axs[fig_index_on_page].plot(xy[:, 0], xy[:, 1], label=label,
                                  linewidth=lw,
                                  color=plot_colors[i])
                else:
                    axs[fig_index_on_page].plot(xy[:, 0], xy[:, 1], label=label,
                                  linewidth=lw)
            if labels:
                axs[fig_index_on_page].annotate(labels[ct], namePos, fontsize=font_size*gen_scaling)
            if annotations:
                axs[fig_index_on_page].annotate(annotations[ct], annotationPos,
                                  fontsize=font_size*gen_scaling)
            if ((print_legend == 'all'
                    or (print_legend == 'first' and fig_index_on_page == 0)
                    or (print_legend == 'tr'
                        and (fig_index_on_page//xfigs == 0 and ((fig_index_on_page+1) % xfigs == 0
                                                  or ct + 1 == n_beams))))
                    and len(data) > 1):
                legendscale = 1.
                if len(data) > 2:
                    legendscale = 1/np.sqrt(len(data)-1)
                legend = axs[fig_index_on_page].legend(fontsize=0.9*font_size*gen_scaling*legendscale,
                                         loc="upper right", frameon=False,
                                         ncol=(len(data) // 3 + 1))
                legend.get_frame().set_linewidth(line_width)
            fig_index_on_page += 1

        # finally, in case the last figure is empty (i.e. the number of beams
        # is odd) turn off the last axes (but leave the blank space).
        if n_beams % figs_per_page != 0:
            for a in axs[-(figs_per_page - n_beams % figs_per_page):]:
                a.axis('off')
    except Exception:
        logger.error("plot_iv: Error while compiling figures.", exc_info=True)
    finally:
        logger.setLevel(loglevel)

    # if a filename is given write to PDF, else return list of figs
    if filename is not None:
        try:
            pdf = PdfPages(filename)
        except PermissionError:
            logger.error(f"writeRfactorPdf: Cannot open file {filename}. "
                         "Aborting.")
            return

        try:
            for fig in figs:
                pdf.savefig(fig, bbox_inches='tight')
                plt.close(fig)
        except Exception:
            logger.error("plot_iv: Error while writing {}: ".format(filename),
                        exc_info=True)
        finally:
            pdf.close()
            return
    else:
        return figs




"""ViPErLEED utility: Strip comments from POSCAR file."""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from viperleed.utilities.poscar.base import _PoscarStreamCLI


class StripCommentsCLI(_PoscarStreamCLI, cli_name='strip_comments'):
    """Remove ViPErLEED (or VASP) comments from a POSCAR file."""

    long_name = 'strip comments'

    def process_slab(self, slab, args):
        """Return an unchanged slab, as we only remove comments."""
        return slab


if __name__ == '__main__':
    StripCommentsCLI.run_as_script()



"""Contains generic functions used in the TensErLEED scripts."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-06-13'
__license__ = 'GPLv3+'

import itertools
import logging
import multiprocessing
import os
import re
import shutil
import subprocess
import sys

import numpy as np
import scipy.spatial as sps

logger = logging.getLogger(__name__)
COLLAPSE_EPS = 1e-8  # Default for collapsing fractional coordinates

###############################################
#                EXCEPTIONS                   #
###############################################

class NonIntegerMatrixError(ValueError):                                         # TODO: move somewhere else
    """A matrix that should have integer values does not."""


class SingularMatrixError(ValueError, ZeroDivisionError):
    """A matrix that needs inversion is singular."""


###############################################
#                 CLASSES                     #
###############################################

class BackwardsReader:                                                          # TODO: maybe better to give it __enter__ and __exit__?
    """Simple class for reading a large file in reverse without having to read
    the entire file to memory.
    From http://code.activestate.com/recipes/120686/, adapted for python 3 and
    changed somewhat to fit requirements.
    """

    def readline(self):
        while len(self.data) == 1 and ((self.blkcount * self.blksize)
                                       < self.size):
            self.blkcount = self.blkcount + 1
            line = self.data[0]
            try:
                # read from end of file
                self.f.seek(-self.blksize * self.blkcount, 2)
                self.data = ((self.f.read(self.blksize).decode(self.encoding))
                             + line).splitlines()
            except IOError:    # can't seek before the beginning of the file
                self.f.seek(0)
                self.data = (((self.f.read(self.size - (self.blksize
                                                        * (self.blkcount-1))))
                              .decode(self.encoding) + line).splitlines())

        if len(self.data) == 0:
            return ""

        line = self.data.pop()
        return line + '\n'

    def close(self):
        self.f.close()

    def __init__(self, file, blksize=4096, encoding="utf-8"):
        """initialize the internal structures"""
        # get the file size
        self.size = os.stat(file)[6]
        # how big of a block to read from the file...
        self.blksize = blksize
        # how many blocks we've read
        self.blkcount = 1
        # what encoding to use
        self.encoding = encoding
        self.f = open(file, 'rb')
        # if the file is smaller than the blocksize, read a block,
        # otherwise, read the whole thing...
        if self.size > self.blksize:
            # read from end of file
            self.f.seek(-self.blksize * self.blkcount, 2)
        self.data = ((self.f.read(self.blksize))
                     .decode(self.encoding).splitlines())
        if not self.data:  # File is empty
            return
        # strip the last item if it's empty... a byproduct of the last line
        # having a newline at the end of it
        if not self.data[-1]:
            self.data.pop()


class CustomLogFormatter(logging.Formatter):
    """Logging Formatter for level-dependent message formatting"""
    FORMATS = {
        logging.DEBUG: "dbg: %(msg)s",
        logging.INFO: "%(msg)s",
        logging.WARNING: "# WARNING: %(msg)s",
        logging.ERROR: "### ERROR ### in %(module)s:%(funcName)s:%(lineno)s\n"
                       "# %(msg)s \n#############",
        logging.CRITICAL: "### CRITICAL ### in %(module)s:%(funcName)s:"
                          "%(lineno)s\n# %(msg)s \n################",
        "DEFAULT": "%(msg)s",
    }

    def format(self, record):
        # debug log format for everything at DEBUG level or lower
        if record.levelno < logging.DEBUG:
            log_fmt = self.FORMATS.get(logging.DEBUG)
        else:
            log_fmt = self.FORMATS.get(record.levelno, self.FORMATS['DEFAULT'])
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


###############################################
#                FUNCTIONS                    #
###############################################

def add_edges_and_corners(cartesian, fractional, releps, ucell, props=None):
    """Add atoms at edges and corners to the Cartesian coordinates given.

    Notice that edges and corners will be added only for dimensions
    up to the smallest among: shape(fractional)[1], len(releps),
    len(ucell). The only requirements are that `fractional` contains
    at most the same number of elements as `cartesian`, and that the
    second axis of `cartesian` and `ucell` are broadcastable. This
    allows, e.g., to pass in 3D coordinates but only add repeats
    for 2D edges and corners.

    Parameters
    ----------
    cartesian : Sequence
        The Cartesian coordinates to which replicas of atoms at
        edges and corners will be appended. Shape (N, 2) or (N, 3).
    fractional : Sequence
        Fractional coordinates corresponding to `cartesian`. Should
        have shape (M, m), with M <= N.
    releps : Sequence
        Tolerance on fractional coordinates to consider an atom from
        `cartesian` to be close to an edge. Shape (2,) or (3,).
    ucell : Sequence
        Unit cell to be used for the displacement of atoms close to
        edges/corners. Shape (2, 2) or (3, 3), and such that unit
        cell vectors are rows. The second dimension should match
        the second one of `cartesian`.
    props : list or None, optional
        If given, it should have the same length as `cartesian`. It
        is interpreted as a list of properties that should also be
        duplicated whenever an atom is duplicated. Default is None.

    Returns
    -------
    cartesian_extended : numpy.ndarray
        Cartesian atomic coordinates to which extra atoms have been
        appended at the end for all those atoms in `cartesian` that
        were at edges/corners.
    props_extended : list or None
        Corresponding properties of the added atoms. None if `props`
        was not given to begin with.
    """
    for i, pos in enumerate(fractional):
        addlist = []
        for posj, epsj, uvec in zip(pos, releps, ucell):
            if abs(posj) < epsj:        # 'left' edge
                addlist.append(cartesian[i] + uvec)
            elif abs(posj - 1) < epsj:  # 'right' edge
                addlist.append(cartesian[i] - uvec)
        if len(addlist) == 2:
            # 2D corner - add the diagonally opposed one
            addlist.append(sum(addlist) - cartesian[i])
        elif len(addlist) == 3:
            # 3D corner - add all diagonally opposed points
            addlist.extend(
                p1 + p2 - cartesian[i]
                for p1, p2 in itertools.combinations(addlist, 2)
                )
            addlist.append(sum(addlist[:3]) - 2 * cartesian[i])

        if not addlist:
            continue
        cartesian = np.concatenate((cartesian, addlist))
        if props is not None:
            props.extend([props[i]]*len(addlist))
    return cartesian, props


def collapse(cartesians, ucell, ucell_inv=None, method='floor'):
    """Collapse Cartesian coordinates to the base cell.

    Parameters
    ----------
    cartesians : numpy.ndarray
        Cartesian coordinates to be collapsed. Shape (N, 2) or (N, 3).
    ucell : numpy.ndarray
        Basis vectors (rows) of the unit cell to be used for collapsing.
    ucell_inv : numpy.ndarray or None, optional
        The inverse of ucell. If not given or None, it is computed
        on the fly. Default is None.
    method : {'floor', 'round'}, optional
        Which method should be used for collapsing. 'floor'
        returns Cartesian coordinates collapsed to the (0, 0)
        cell, i.e., with all fractional coordinates in range
        [0, 1]. 'round' collapses the coordinates so that they
        are as close to zero as possible. Fractional coordinates
        are then in range [-0.5, 0.5].

    Returns
    -------
    collapsed_cartesians : numpy.ndarray
        The Cartesian coordinates, collapsed according to method
    fractional_coordinates : numpy.ndarray
        The corresponding fractional coordinates
    """
    if ucell_inv is None:
        ucell_inv = np.linalg.inv(ucell)
    fractional = collapse_fractional(cartesians.dot(ucell_inv),
                                     method=method)
    return fractional.dot(ucell), fractional


def _floor_eps(eps):
    """Return the floored-int version of values after adding eps."""
    def _floor(values):
        return np.floor(values + eps)
    return _floor


def collapse_fractional(coordinates, method='floor',
                        eps=COLLAPSE_EPS, in_place=False):
    """Collapse fractional coordinates to the base cell.

    Parameters
    ----------
    coordinates : numpy.ndarray
        Fractional coordinates to be collapsed. Shape (N, 2) or (N, 3).
    method : {'floor', 'round'}, optional
        Which method should be used for collapsing.  'round' collapses
        the coordinates so that they are as close to zero as possible.
        Fractional coordinates are then in range [-0.5, 0.5]. 'floor'
        returns fractional coordinates collapsed to the (0, 0) cell,
        that is, with all fractional coordinates in range [-`eps`,
        1-`eps`]. This ensures that collapsing a fractional coordinate
        of (essentially) 1 gives (essentially) zero. Default is 'floor'.
    eps : float or Sequence, optional
        Used only if method == 'floor'. The (fractional) tolerance
        for collapsing. If a sequence, it should have as many items
        as the second axis of coordinates. Default is 1e-8.
    in_place : bool, optional
        If True, the function modifies directly `coordinates`. Can be
        used to save some memory. Default is False.

    Returns
    -------
    collapsed : numpy.ndarray
        The collapsed fractional coordinates. Will be the same object
        passed in if in_place is True-thy.

    Raises
    ------
    ValueError
        If method is not one of the valid methods.
    """
    _methods = {'f': _floor_eps(eps), 'r': np.round}
    try:
        round_ = _methods[method[0]]
    except (TypeError, IndexError, KeyError):
        raise ValueError(f'collapse_fractional: Unknown method={method}')
    if in_place:
        collapsed = coordinates
        collapsed -= round_(coordinates)
    else:
        collapsed = coordinates - round_(coordinates)
    return collapsed


def ensure_integer_matrix(matrix, eps=1e-6):
    """Return a rounded version of matrix. Raise if matrix isn't integer."""
    rounded = np.round(matrix)
    if np.any(abs(matrix - rounded) > eps):
        raise NonIntegerMatrixError(matrix)
    return rounded


def pairwise(iterable):
    """Return a generator of pairs of subsequent elements."""
    try:
        # pairwise is available in py >= 3.10
        return itertools.pairwise(iterable)
    except AttributeError:
        pass
    # Rough equivalent from the python docs
    orig, offset = itertools.tee(iterable)
    next(offset, None)
    return zip(orig, offset)


def rotation_matrix(angle, dim=2):
    """Returns a (2x2) matrix for in-plane rotation of the given rotation
    angle. Set dim=3 to get a 3x3 matrix with rotation in [:2, :2]."""
    if dim < 2:
        raise ValueError('Rotation matrix needs at least dimension 2')
    m = np.eye(dim, dtype=float)
    m[:2, :2] = np.array([[np.cos(angle), -np.sin(angle)],
                          [np.sin(angle), np.cos(angle)]])
    return m


def rotation_matrix_order(order, dim=2):
    """Returns a (2x2) matrix for in-plane rotation of the given rotation
    order. Set dim=3 to get a 3x3 matrix with rotation in [:2, :2]."""
    angle = 2*np.pi/order
    return rotation_matrix(angle, dim=dim)


def fortranContLine(s):
    """Takes a sting that might be too long to fit on a single line of fortran
    code and splits it into continuation lines as necessary. Returns a string
    with ampersands and line breaks."""
    limit = 72  # fortran line length limit
    if len(s) <= limit:
        return s
    o = s[:6]
    s = s[6:]
    while len(s) > (limit-6):   # 6 characters for beginning of line
        o += s[:(limit-6)]
        o += "&\n     &"
        s = s[(limit-6):]
    o += s
    return o


def get_elapsed_time_str(t):
    """
    Takes float time in seconds, returns a formatted string giving the elapsed
    times in minutes, hours or seconds, as appropriate.
    """
    if t >= 3600:
        elapsedTimeStr = (str(int(t/3600)) + ":"+(str(int(t/60) % 60)).zfill(2)
                          + " hours")
    elif t >= 60:
        elapsedTimeStr = (str(int(t/60)) + ":"+(str(int(t) % 60)).zfill(2)
                          + " minutes")
    else:
        elapsedTimeStr = str(round(t, 2)) + " seconds"
    return elapsedTimeStr


def readIntRange(s):
    """Takes a string, returns a list of integers. If the string is a single
    int or space-separated ints, the return value is a list containing only
    those int. If the string contains i1-i2 or i1:i2, the list is
    range(i1, i2+1), i.e. contains both i1 and i2."""
    out = []
    sl = s.split()
    for ss in sl:
        try:
            out.append(int(ss))
        except ValueError:
            try:
                start, stop = split_string_range(ss)
            except ValueError:
                return []

            try:
                out.extend(range(int(start), int(stop)+1))
            except ValueError:
                return []
    return list(dict.fromkeys(out))  # Like set, but keep order


def split_string_range(range_string):
    """Return start and stop as strings from "start:stop" or "start-stop"."""
    range_list = []
    if '-' in range_string:
        range_list = range_string.split('-')
    elif ':' in range_string:
        range_list = range_string.split(':')
    try:
        return range_list[0], range_list[-1]
    except IndexError:
        raise ValueError(f"Invalid range string: {range_string}") from None


def range_to_str(il):
    """Takes a list of integers, sorts them and returns a string short form.
    For example, [1, 6, 4, 5, 2, 8] will return "1-2, 4-6, 8". Double entries
    will be ignored."""
    if not all(isinstance(v, int) for v in il):
        t = next(type(v) for v in il if not isinstance(v, int))
        raise TypeError(
            f"range_to_str: expected list of int, found type {t.__name__}"
            )
    sl = sorted(il, reverse=True)
    prev = sl.pop()
    rmin = prev
    out = str(prev)
    while sl:
        v = sl.pop()
        if v == prev:
            continue
        if v - prev == 1:
            prev = v
            continue
        if prev != rmin:
            out += f"-{prev}, {v}"
        else:
            out += f", {v}"
        prev = v
        rmin = v
    if prev != rmin:
        out += f"-{prev}"
    return out


def readVector(s, ucell=None, defRelaltive=False):
    """Takes a string 'xyz[f1 f2 f3]', 'abc[f1 f2 f3]' or just '[f1 f2 f3]'
    and returns the corresponding vector in cartesian coordinates, or None if
    the string cannot be parsed."""
    m = re.match(r'\s*(xyz|abc)?\[\s*(?P<v1>[-0-9.]+)\s+'
                 r'(?P<v2>[-0-9.]+)\s+(?P<v3>[-0-9.]+)\s*\]', s)
    if not m:
        return None
    try:
        v1 = float(m.group("v1"))
        v2 = float(m.group("v2"))
        v3 = float(m.group("v3"))
    except (ValueError, IndexError):
        return None
    if "abc" in s:
        if ucell is None:
            return None
        uct = ucell.T
        return np.dot((v1, v2, v3), ucell.T)
    # xyz
    return np.array([v1, v2, v3])


def readIntLine(line, width=3):                                                 # TODO: Probably better ways with list comprehension
    """
    Reads an (arbitrary length) line of integers with fixed width. Will try
    to interpret everything as integers until the line ends.

    Parameters
    ----------
    line : str
        The line to interpret
    width : integer, optional
        The width of each integer. The default is 3.

    Returns
    -------
    Tuple of integers
    """
    line = line.rstrip()
    out = []
    while line:
        chunk, line = line[:width], line[width:]
        out.append(int(chunk))
    return tuple(out)


def cosvec(x, y):
    """
    Returns the cosine of the angle between two vectors

    Parameters
    ----------
    x : numpy.array
        First vector
    y : numpy.array
        Second vector

    Returns
    -------
    float
        Cosine of the angle between the two vectors

    """
    return np.dot(x, y) / (np.linalg.norm(x) * np.linalg.norm(y))


def dict_equal(d1, d2):                                                         # TODO: d1 == d2 works the same
    """
    Checks whether two dictionaries are equal, i.e. contain the same set of
    keys with the same values

    Parameters
    ----------
    d1 : dict
        First dictionary
    d2 : dict
        Second dictionary

    Returns
    -------
    bool
        True if all keys and values match, False otherwise
    """
    if len({k: d1[k] for k in d1 if k in d2 and d1[k] == d2[k]})-len(d1) == 0:
        return True
    return False


def lcm(a, b):
    "Calculate the lowest common multiple of two integers a and b"
    return a * b // np.gcd(a, b)


def max_diff(list_):
    """Find the index in a list such that list[i] - list[i-1] is the largest
    value difference in the list (after sorting). Return i and the difference.
    """
    if len(list_) < 2:
        return None, None
    list_ = sorted(list_)
    maxdiff = max([list_[i] - list_[i-1] for i in range(1, len(list_))])
    m = [i for i in range(1, len(list_)) if list_[i] - list_[i-1] == maxdiff]
    if m:
        return m[0], maxdiff
    return None, maxdiff


def parseMathSqrt(s):
    try:
        f = float(s)
    except ValueError:
        f = 1.0
        if '*' in s:
            sl = s.split('*')
        else:
            sl = [s]
        sl2 = []
        for el in sl:
            if '/' in el:
                sl2.append(el.split('/')[0])
                for s in el.split('/')[1:]:
                    sl2.append(1/parseMathSqrt(s))
            else:
                sl2.append(el)
        for el in sl2:
            try:
                f *= float(el)
            except ValueError:
                if 'sqrt' not in el:
                    logger.error('Could not interpret '+el+' as float value')
                    raise
                else:
                    p = re.compile(r'\s*sqrt\s*\(\s*(?P<value>[\d.]*)\s*\)')
                    m = p.match(el)
                    if m:
                        try:
                            f *= np.sqrt(float(m.group('value')))
                        except ValueError:
                            logger.error('Could not interpret ' +
                                         m.group('value')+' as float value')
                            raise
                    else:
                        logger.error('Could not interpret ' + el
                                     + ' as float value')
                        raise
    return f

# def angle(v1, v2, acute=True):
#     """Returns the angle between two vectors"""
#  !! UNSIGNED!
#     angle = np.arccos(np.dot(v1, v2)
#                       / (np.linalg.norm(v1)*np.linalg.norm(v2)))
#     if acute == True:
#         return angle
#     else:
#         return 2 * np.pi - angle


def angle(v1, v2):
    """Returns the angle between two 2D vectors"""
    # Use cross product for sine, dot product for cosine
    return np.arctan2(v1[0]*v2[1] - v1[1]*v2[0], v1[0]*v2[0] + v1[1]*v2[1])


def dist_from_line(p1, p2, r):
    """
    Calculates the distance of a point from a line, with the line defined by
    two other points. Works in both 2 and 3 dimensions.

    Parameters
    ----------
    p1, p2 : np.arrays of size 2 or 3
        The points defining the line.
    r : np.array of same size as p1, p2
        The point for which the distance should be determined.

    Returns
    -------
    float
        The distance.
    """
    if len(p1) == 2:
        return (abs((p2[1] - p1[1]) * r[0] - (p2[0] - p1[0]) * r[1]
                    + p2[0] * p1[1] - p2[1] * p1[0])
                / np.sqrt((p2[1] - p1[1])**2 + (p2[0] - p1[0])**2))
    if len(p1) == 3:
        return (np.linalg.norm(np.cross((r - p1), (p2 - p1)))
                / np.linalg.norm(p2 - p1))
    raise ValueError("Vector dimensions have to be either 2 or 3.")


def parent_name(dotted_name, remove=''):
    """Return a version of dotted_name with the last attribute removed.

    Parameters
    ----------
    dotted_name : str
        The dotted name (e.g., the name of an attribute or a module).
    remove : str, optional
        What should be removed from the right side. It may, or may
        not contain a dot. If it does not begin with a leading dot,
        a single dot is prepended to `remove`. If not given, the
        last dotted portion is removed. Default is an empty string.

    Returns
    -------
    stripped_name : str
        A version of dotted_name with [.]remove removed from the right.
    """
    remove = remove or '.'
    if not remove.startswith('.'):
        remove = '.' + remove
    stripped_name, *_ = dotted_name.rsplit(remove, maxsplit=1)
    return stripped_name


def strip_comments(line):
    """Return the part of line to the left of comments."""
    for comment_char in "!#%":
        try:
            line, *_ = line.split(comment_char)
        except ValueError:  # Nothing left to split
            return ''
    return line.strip()


def readToExc(llist):                                                           # TODO: unused; could be an iterator
    """For reading PARAMETERS files; takes a list, returns elements until the
    first one that starts with an exclamation mark."""
    read = True
    newlist = []
    for s in llist:
        if read:
            if s[0] == '!':
                read = False
            else:
                newlist.append(s)
    return newlist


def splitSublists(llist, sep):                                                  # TODO: could be an iterator
    """Takes a list and a separator, splits strings in the list by the
    separator, returns results as list of lists"""
    newlist = []
    sublist = []
    for el in llist:
        if sep not in el:
            sublist.append(el)
        else:
            pl = el.split(sep)
            if pl[0]:
                sublist.append(pl[0])
            newlist.append(sublist)
            sublist = []
            if len(pl) > 1:
                for i in range(1, len(pl)):
                    s = pl[i]
                    if s:
                        sublist.append(s)
                    if i != len(pl) - 1:
                        newlist.append(sublist)
                        sublist = []
    newlist.append(sublist)
    return(newlist)


def splitMaxRight(s, sep):
    """Same as s.split(sep, maxsplit=1), but splitting at the first instance
    from the right."""
    sr = s[::-1]
    L = sr.split(sep, maxsplit=1)
    L.reverse()
    nl = []
    for ns in L:
        nl.append(ns[::-1])
    return nl


def recombineListElements(llist, com):
    """Takes a list, checks in each element whether the first/last characters
    are the given combination character, and if so, combines list elements with
    the list element before/after."""
    i = 0
    newlist = llist[:]
    while i < len(newlist)-1:
        if newlist[i][-1] == com or newlist[i+1][0] == com:
            newlist[i] += newlist.pop(i+1)
        else:
            i += 1
    return newlist


def addUnequalPoints(l1, l2, eps, uniqueLists=False):
    """Adds all points from l1 to l2, if they are not already in l2
    (+- epsilon)."""
    nl2 = l2[:]
    nl1 = l1[:]
    if len(l2) == 0:
        nl2 = nl1
    else:
        if not uniqueLists:
            # first get rid of duplicates in nl1
            tree = sps.KDTree(nl1)
            usepoint = [True] * len(nl1)
            for (i, p) in enumerate(nl1):
                if usepoint[i]:
                    for j in tree.query_ball_point(p, eps)[1:]:
                        usepoint[j] = False
            nl1 = list(itertools.compress(nl1, usepoint))
        # then add remaining elements to l2:
        dl = sps.distance.cdist(np.vstack(tuple(l1)), np.vstack(tuple(l2)),
                                'euclidean')
        for (i, sublist) in enumerate(dl):
            if min(sublist) >= eps:
                nl2.append(l1[i])
    return nl2


def available_cpu_count():
    """ Number of available virtual or physical CPUs on this system, i.e.
    user/real as output by time(1) when called with an optimally scaling
    userspace-only program"""

    # cpuset
    # cpuset may restrict the number of *available* processors
    try:
        with open('/proc/self/status') as f:
            m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$', f.read())
        if m:
            res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
            if res > 0:
                return res
    except IOError:
        pass

    # Python 2.6+
    try:
        return multiprocessing.cpu_count()
    except NotImplementedError:
        pass

    # POSIX
    try:
        res = int(os.sysconf('SC_NPROCESSORS_ONLN'))

        if res > 0:
            return res
    except (AttributeError, ValueError):
        pass

    # Windows
    try:
        res = int(os.environ['NUMBER_OF_PROCESSORS'])
        if res > 0:
            return res
    except (KeyError, ValueError):
        pass

    # BSD
    try:
        sysctl = subprocess.Popen(['sysctl', '-n', 'hw.ncpu'],
                                  stdout=subprocess.PIPE)
        scStdout = sysctl.communicate()[0]
        res = int(scStdout)
        if res > 0:
            return res
    except (OSError, ValueError):
        pass

    # Linux
    try:
        with open('/proc/cpuinfo') as f:
            res = f.read().count('processor\t:')
        if res > 0:
            return res
    except IOError:
        pass

    # Solaris
    try:
        pseudoDevices = os.listdir('/devices/pseudo/')
        res = 0
        for pd in pseudoDevices:
            if re.match(r'^cpuid@[0-9]+$', pd):
                res += 1
        if res > 0:
            return res
    except OSError:
        pass

    # Other UNIXes (heuristic)
    try:
        try:
            with open('/var/run/dmesg.boot') as f:
                dmesg = f.read()
        except IOError:
            dmesgProcess = subprocess.Popen(['dmesg'], stdout=subprocess.PIPE)
            dmesg = dmesgProcess.communicate()[0]
        res = 0
        while '\ncpu' + str(res) + ':' in dmesg:
            res += 1
        if res > 0:
            return res
    except OSError:
        pass

    return -1


def copytree_exists_ok(source, destination):
    """Copy the whole tree at the `source` directory to `destination`.

    This is a wrapper around the shutil.copytree function that
    maintains backwards compatibility down to python v3.5.

    Parameters
    ----------
    source : Path
        Base of the directory tree to be copied. Notice that symlinks
        in `source` will NOT be handled correctly for python < 3.8.
    destination : Path
        Path to the directory that will mirror source and its contents.
        It is created if it does not exist yet.

    Returns
    -------
    None.
    """
    if sys.version_info >= (3, 8):
        # dirs_exist_ok was introduced in python 3.8:
        # https://docs.python.org/3/library/shutil.html#shutil.copytree
        # pylint: disable-next=unexpected-keyword-arg
        shutil.copytree(source, destination, dirs_exist_ok=True)
        return
    # For earlier python versions, we need to do things manually. We
    # use a simplified version of the implementation of copytree from
    # shutil for py3.8. We assume that source and destination are Path
    # objects, and that we don't have anything special like symlinks.
    # The next line will not work in py<3.5 because of exist_ok.
    destination.mkdir(parents=True, exist_ok=True)
    for srcentry in source.glob('*'):
        dstentry = destination / srcentry.name
        if srcentry.is_dir():
            copytree_exists_ok(srcentry, dstentry)
        else:  # file
            shutil.copy2(srcentry, dstentry)
    shutil.copystat(source, destination)


def make_unique_list(w_duplicates):                                             # TODO: better function in guilib.helpers
    """Helper function to remove duplicates from list. Does same as creating a set but preservers order.

    Args:
        w_duplicates (iterable): list with duplicates

    Returns:
        list: list with duplictes removed
    """
    unique_list = []
    for item in w_duplicates:
        if item not in unique_list:
            unique_list.append(item)
    return unique_list



"""Module iotensors of viperleed.calc.files.

Defines functionality useful for reading information from Tensor files.

Part of the functionality in this module used to be part of leedbase
(originally 2019-06-13). Moved here to reduce cyclic import issues.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-22'
__license__ = 'GPLv3+'

import copy
import logging
import os
from pathlib import Path
import re
import shutil
from zipfile import ZipFile

from viperleed.calc.files import parameters, poscar, vibrocc

logger = logging.getLogger(__name__)


def getTensors(index, base_dir=".", target_dir=".", required=True):
    """Fetches Tensor files from Tensors or archive with specified tensor
    index. If required is set True, an error will be printed if no Tensor
    files are found.
    base_dir is the directory in which the Tensor directory is based.
    target_dir is the directory to which the Tensor files should be moved."""
    dn = "Tensors_"+str(index).zfill(3)
    tensor_dir = (Path(base_dir) / "Tensors").resolve()
    unpack_path = (Path(target_dir) / "Tensors" / dn).resolve()
    zip_path = (tensor_dir / dn).with_suffix(".zip")

    if (os.path.basename(base_dir) == "Tensors"
            and not tensor_dir.is_dir()):
        base_dir = os.path.dirname(base_dir)
    if not (tensor_dir / dn).is_dir():
        if (tensor_dir / dn).with_suffix(".zip").is_file():
            logger.info(f"Unpacking {dn}.zip...")
            os.makedirs(unpack_path, exist_ok=True)
            try:
                with ZipFile(zip_path, 'r') as zip_ref:
                    zip_ref.extractall(unpack_path)                             # TODO: maybe it would be nicer to read directly from the zip file
            except Exception:
                logger.error(f"Failed to unpack {dn}.zip")
                raise
        else:
            logger.error("Tensors not found")
            raise RuntimeError("Tensors not found")
    elif base_dir != target_dir:
        try:
            os.makedirs(unpack_path, exist_ok=True)
            for file in os.listdir(os.path.join(base_dir, "Tensors", dn)):
                shutil.copy2(file, unpack_path)
        except OSError:
            logger.error(f"Failed to move Tensors from {dn}")
            raise


def getTensorOriStates(sl, path):
    """Reads POSCAR, PARAMETERS and VIBROCC from the target path, gets the
    original state of the atoms and sites, and stores them in the given
    slab's atom/site oriState variables."""
    path = Path(path).resolve()
    for fn in ["POSCAR", "PARAMETERS", "VIBROCC"]:
        if not (path / fn).is_file():
            logger.error(f"No {fn} file in {path}")
            raise RuntimeError("Could not check Tensors: File missing")         # TODO: FileNotFoundError?
    dn = path.parent
    try:
        tsl = poscar.read(path / "POSCAR")
        trp = parameters.read(filename=path/"PARAMETERS")
        parameters.interpret(trp, slab=tsl, silent=True)
        tsl.full_update(trp)
        vibrocc.readVIBROCC(trp, tsl, filename=path/"VIBROCC", silent=True)
        tsl.full_update(trp)
    except Exception as exc:
        logger.error("Error checking Tensors: Error while reading "
                     f"input files in {dn}")
        logger.debug("Exception:", exc_info=True)
        raise RuntimeError("Could not check Tensors: Error loading old input "
                           "files") from exc
    if tsl.n_atoms != sl.n_atoms:
        logger.error(f"POSCAR from {dn} is incompatible with "
                     "current POSCAR.")
        raise RuntimeError("Tensors file incompatible")
    for at in sl:
        tat = tsl.atlist.get(at.num, None)
        if tat is None:
            logger.error(f"POSCAR from {dn} is incompatible with "
                         "current POSCAR.")
            raise RuntimeError("Tensors file incompatible")
        at.copyOriState(tat)
    if len(tsl.sitelist) != len(sl.sitelist):
        logger.error(f"Sites from {dn} input differ from current input.")
        raise RuntimeError("Tensors file incompatible")
    for site in sl.sitelist:
        tsitel = [s for s in tsl.sitelist if site.label == s.label]
        if len(tsitel) != 1:
            logger.error(f"Sites from {dn} input differ from current input.")
            raise RuntimeError("Tensors file incompatible")
        site.oriState = copy.deepcopy(tsitel[0])
